# Copyright (c) 2016 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


def trace_response_to_thriftrw(service, tr):
    if tr is None:
        return None
    return service.TraceResponse(span=observed_span_to_thriftrw(service, tr.span),
                                 downstream=trace_response_to_thriftrw(service, tr.downstream),
                                 notImplementedError=tr.notImplementedError or '')


def downstream_to_thriftrw(service, downstream):
    if downstream is None:
        return None
    return service.Downstream(downstream.serviceName,
                              downstream.serverRole,
                              downstream.host,
                              downstream.port,
                              downstream.transport,
                              downstream.downstream)


def join_trace_request_to_thriftrw(service, jtr):
    return service.JoinTraceRequest(jtr.serverRole, downstream_to_thriftrw(service, jtr.downstream))


def observed_span_to_thriftrw(service, observed_span):
    return service.ObservedSpan(traceId=observed_span.traceId,
                                sampled=observed_span.sampled,
                                baggage=observed_span.baggage)



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import

import time

import pytest

from jaeger_client.ioloop_util import PeriodicCallback
from .conftest import wait_for


class TestPeriodicCallback(object):

    def test_callback_and_callback_time_required(self):
        with pytest.raises(TypeError):
            PeriodicCallback()

        with pytest.raises(TypeError):
            PeriodicCallback(callback=lambda: True)

        with pytest.raises(TypeError):
            PeriodicCallback(callback_time=100)

    def test_callback_time_gt_zero(self):
        with pytest.raises(ValueError) as exc:
            PeriodicCallback(lambda: True, -1)
        assert str(exc.value) == 'callback_time must be positive'

    def test_start_runs_callback_and_stop_halts_scheduling(self):
        count = [0]

        def callback():
            count[0] += 1

        pc = PeriodicCallback(callback, 500)
        assert not pc._running
        assert pc._scheduler_thread is None
        assert pc._callback_thread is None

        pc.start()
        assert wait_for(lambda: count[0])
        pc.stop()

        assert count[0] == 1
        assert pc._running is False
        assert pc._scheduler_thread is None
        assert pc._callback_thread is None

    def test_runtime_exceeding_period_prevents_subsequent_runs(self):
        count = [0]

        def callback():
            time.sleep(.05)
            count[0] += 1
            time.sleep(.05)

        pc = PeriodicCallback(callback, 1)
        pc.start()
        assert wait_for(lambda: count[0])
        pc.stop()

        assert count[0] == 1
        assert pc._running is False
        assert pc._scheduler_thread is None
        assert pc._callback_thread is None

    def test_start_is_idempotent(self):
        count = [0]

        def callback():
            time.sleep(.05)
            count[0] += 1
            time.sleep(.05)

        pc = PeriodicCallback(callback, 1)
        for _ in range(100):
            pc.start()

        assert wait_for(lambda: count[0])
        pc.stop()

        assert count[0] == 1
        assert pc._running is False
        assert pc._scheduler_thread is None
        assert pc._callback_thread is None

    def test_out_of_order_invocations_harmless(self):
        count = [0]

        def callback():
            count[0] += 1
            time.sleep(.05)

        pc = PeriodicCallback(callback, 0)

        pc.stop()
        pc.start()
        assert wait_for(lambda: count[0])
        pc.stop()

        assert count[0] == 1
        assert pc._running is False
        assert pc._scheduler_thread is None
        assert pc._callback_thread is None



# Copyright (c) 2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import print_function

import sys
import time
import socket
import collections

import mock
import requests
import requests.auth
import pytest

from jaeger_client import senders, thrift
from jaeger_client import Span, SpanContext
from jaeger_client.local_agent_net import LocalAgentSender
from jaeger_client.thrift_gen.agent import Agent
from jaeger_client.thrift_gen.jaeger import ttypes
from thrift.protocol import TCompactProtocol


def test_base_sender_send_not_implemented():
    sender = senders.Sender()
    with pytest.raises(NotImplementedError):
        sender.send(1)


def test_base_sender_set_process_instantiate_jaeger_process():
    sender = senders.Sender()
    sender.set_process('service', {}, max_length=0)
    assert isinstance(sender._process, ttypes.Process)
    assert sender._process.serviceName == 'service'


def test_base_sender_spanless_flush_is_noop():
    sender = senders.Sender()
    flushed = sender.flush()
    assert flushed == 0


def test_base_sender_processless_flush_is_noop():
    sender = senders.Sender()
    sender.spans.append('foo')
    flushed = sender.flush()
    assert flushed == 0


class CustomException(Exception):
    pass


@pytest.fixture
def channel():
    channel = mock.MagicMock()
    channel._reporting_port = 4242
    channel._host = 'mock'
    return channel


class TestSenderFlush(object):

    def span(self):
        FakeTracer = collections.namedtuple(
            'FakeTracer',
            ['ip_address', 'service_name', 'max_tag_value_length']
        )
        tracer = FakeTracer(
            ip_address='127.0.0.1',
            service_name='reporter_test',
            max_tag_value_length=sys.maxsize
        )
        ctx = SpanContext(trace_id=1, span_id=1, parent_id=None, flags=1)
        span = Span(context=ctx, tracer=tracer, operation_name='foo')
        span.start_time = time.time()
        span.end_time = span.start_time + 0.001  # 1ms
        return span

    def test_base_sender_flush_raises_exceptions(self):
        sender = senders.Sender()
        sender.set_process('service', {}, max_length=0)

        sender.spans = [self.span()]

        sender.send = mock.MagicMock(side_effect=CustomException('Failed to send batch.'))
        assert sender.span_count == 1

        try:
            sender.flush()
        except Exception as exc:
            assert isinstance(exc, CustomException)
            assert str(exc) == 'Failed to send batch.'
        else:
            assert False, "Didn't Raise"
        assert sender.span_count == 0

    def test_udp_sender_flush_reraises_exceptions(self, channel):
        exceptions = ((CustomException, 'Failed to send batch.',
                       'Failed to submit traces to jaeger-agent: Failed to send batch.'),
                      (socket.error, 'Connection Failed',
                       'Failed to submit traces to jaeger-agent socket: Connection Failed'))
        for exception, value, expected_value in exceptions:
            sender = senders.UDPSender(channel)
            sender.set_process('service', {}, max_length=0)

            sender.spans = [self.span()]

            def mock_fetch(*args):
                raise exception(value)

            assert sender.span_count == 1

            with mock.patch.object(sender._agent, 'emitBatch', mock_fetch):
                try:
                    sender.flush()
                except Exception as exc:
                    assert isinstance(exc, exception)
                    assert str(exc) == expected_value
                else:
                    assert False, "Didn't Raise"
                assert sender.span_count == 0

    def test_udp_sender_large_span_dropped_on_flush(self, channel):
        sender = senders.UDPSender(channel)
        sender.set_process('service', {'tagOne': 'someTagValue'}, max_length=0)

        span_to_send = self.span()
        span_to_filter = self.span()
        span_to_filter.set_tag('someTag', '.' * 65000)
        sender.spans = [span_to_send, span_to_filter]

        batch_store = []

        def mock_emit_batch(batch):
            batch_store.append(batch)

        with mock.patch.object(sender._agent, 'emitBatch', mock_emit_batch):
            try:
                sender.flush()
            except Exception as exc:
                assert isinstance(exc, senders.UDPSenderException)
                assert 'Cannot send span of size' in str(exc)
            else:
                assert False, "Didn't Raise"

        assert batch_store[0].spans == [thrift.make_jaeger_span(span_to_send)]

    def test_udp_sender_batches_spans_over_multiple_packets_on_flush(self, channel):
        sender = senders.UDPSender(channel)
        sender.set_process('service', {'tagOne': 'someTagValue'}, max_length=0)

        spans = [self.span() for _ in range(60)]
        for span in spans:
            span.set_tag('tag', '.' * 10000)  # Send 10 batches of 6
        sender.spans = list(spans)

        batch_store = []

        def mock_emit_batch(batch):
            batch_store.append(batch)

        with mock.patch.object(sender._agent, 'emitBatch', mock_emit_batch):
            sender.flush()

        assert len(batch_store) == 10
        c = 0
        for i in range(10):
            assert batch_store[i].spans == [
                thrift.make_jaeger_span(span) for span in spans[c:c + 6]
            ]
            c += 6


def test_udp_sender_instantiate_thrift_agent(channel):
    sender = senders.UDPSender(channel)

    assert sender._agent is not None
    assert isinstance(sender._agent, Agent.Client)


def test_udp_sender_intantiate_local_agent_channel():
    sender = senders.UDPSender()

    assert sender._channel is not None
    assert isinstance(sender._channel, LocalAgentSender)


def test_udp_sender_calls_agent_emitBatch_on_send(channel):
    test_data = {'foo': 'bar'}
    sender = senders.UDPSender(channel)
    sender._agent = mock.Mock()
    sender.send(test_data)
    sender._agent.emitBatch.assert_called_once_with(test_data)


def test_udp_sender_implements_thrift_protocol_factory(channel):
    sender = senders.UDPSender(channel)

    assert callable(sender.getProtocol)
    protocol = sender.getProtocol(mock.MagicMock())
    assert isinstance(protocol, TCompactProtocol.TCompactProtocol)


class TestHTTPSender(object):

    def loaded_sender(self, *args, **kwargs):
        FakeTracer = collections.namedtuple('FakeTracer', ['ip_address', 'service_name'])
        tracer = FakeTracer(ip_address='127.0.0.1', service_name='reporter_test')
        ctx = SpanContext(trace_id=1, span_id=1, parent_id=None, flags=1)
        span = Span(context=ctx, tracer=tracer, operation_name='foo')
        span.start_time = time.time()
        span.end_time = span.start_time + 0.001
        sender = senders.HTTPSender(*args, **kwargs)
        sender.set_process('service', {}, max_length=0)
        sender.spans = [span]
        return sender

    def test_http_sender_provides_proper_request_content(self):
        sender = self.loaded_sender('http://some_endpoint')

        with mock.patch('requests.sessions.Session.post') as post:
            sender.flush()
            time.sleep(0.001)
            kwargs = post.call_args[1]
            assert kwargs['url'] == 'http://some_endpoint'
            assert kwargs['headers'].get('Content-Type') == 'application/x-thrift'
            assert int(kwargs['headers'].get('Content-Length')) == len(kwargs['data'])
            assert kwargs['headers'].get('Authorization') is None

    def test_http_sender_provides_auth_token(self):
        sender = self.loaded_sender('http://some_endpoint', auth_token='SomeAuthToken')

        with mock.patch('requests.sessions.Session.send') as send:
            sender.flush()
            request = send.call_args[0][0]
            assert request.headers.get('Authorization') == 'Bearer SomeAuthToken'

    def test_http_sender_provides_basic_auth(self):
        sender = self.loaded_sender('http://some_endpoint',
                                    user='SomeUser', password='SomePassword')

        mock_request = mock.MagicMock()
        mock_request.headers = {}
        requests.auth.HTTPBasicAuth('SomeUser', 'SomePassword')(mock_request)
        expected_auth_header = mock_request.headers['Authorization']

        with mock.patch('requests.sessions.Session.send') as send:
            sender.flush()
            time.sleep(0.001)
            request = send.call_args[0][0]
            assert request.headers['Authorization']
            assert request.headers['Authorization'] == expected_auth_header

    def test_http_sender_flush_reraises_exceptions(self):
        exceptions = ((CustomException('Failed to send batch.'),
                       'POST to jaeger_endpoint failed: Failed to send batch.'),
                      (requests.HTTPError(500, 'Server Error'),
                      'POST to jaeger_endpoint failed: [Errno 500] Server Error'))

        for exception, expected_value in exceptions:
            sender = self.loaded_sender('http://some_endpoint')

            def send(*args, **kwargs):
                raise exception

            with mock.patch('requests.sessions.Session.send', send):
                with pytest.raises(type(exception)) as exc:
                    sender.flush()
                assert str(exc.value) == expected_value
                assert sender.span_count == 0



# Modified by SignalFx
#
# Autogenerated by Thrift Compiler (0.9.3)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:new_style
#
import six
from six.moves import xrange

from thrift.Thrift import TType, TMessageType, TException, TApplicationException

from thrift.transport import TTransport
from thrift.protocol import TBinaryProtocol, TProtocol
try:
  from thrift.protocol import fastbinary
except:
  fastbinary = None


class AnnotationType(object):
  BOOL = 0
  BYTES = 1
  I16 = 2
  I32 = 3
  I64 = 4
  DOUBLE = 5
  STRING = 6

  _VALUES_TO_NAMES = {
    0: "BOOL",
    1: "BYTES",
    2: "I16",
    3: "I32",
    4: "I64",
    5: "DOUBLE",
    6: "STRING",
  }

  _NAMES_TO_VALUES = {
    "BOOL": 0,
    "BYTES": 1,
    "I16": 2,
    "I32": 3,
    "I64": 4,
    "DOUBLE": 5,
    "STRING": 6,
  }


class Endpoint(object):
  """
  Indicates the network context of a service recording an annotation with two
  exceptions.

  When a BinaryAnnotation, and key is CLIENT_ADDR or SERVER_ADDR,
  the endpoint indicates the source or destination of an RPC. This exception
  allows zipkin to display network context of uninstrumented services, or
  clients such as web browsers.

  Attributes:
   - ipv4: IPv4 host address packed into 4 bytes.

  Ex for the ip 1.2.3.4, it would be (1 << 24) | (2 << 16) | (3 << 8) | 4
   - port: IPv4 port

  Note: this is to be treated as an unsigned integer, so watch for negatives.

  Conventionally, when the port isn't known, port = 0.
   - service_name: Service name in lowercase, such as "memcache" or "zipkin-web"

  Conventionally, when the service name isn't known, service_name = "unknown".
   - ipv6: IPv6 host address packed into 16 bytes. Ex Inet6Address.getBytes()
  """

  thrift_spec = (
    None, # 0
    (1, TType.I32, 'ipv4', None, None, ), # 1
    (2, TType.I16, 'port', None, None, ), # 2
    (3, TType.STRING, 'service_name', None, None, ), # 3
    (4, TType.STRING, 'ipv6', None, None, ), # 4
  )

  def __init__(self, ipv4=None, port=None, service_name=None, ipv6=None,):
    self.ipv4 = ipv4
    self.port = port
    self.service_name = service_name
    self.ipv6 = ipv6

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.I32:
          self.ipv4 = iprot.readI32()
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.I16:
          self.port = iprot.readI16()
        else:
          iprot.skip(ftype)
      elif fid == 3:
        if ftype == TType.STRING:
          self.service_name = iprot.readString()
        else:
          iprot.skip(ftype)
      elif fid == 4:
        if ftype == TType.STRING:
          self.ipv6 = iprot.readString()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('Endpoint')
    if self.ipv4 is not None:
      oprot.writeFieldBegin('ipv4', TType.I32, 1)
      oprot.writeI32(self.ipv4)
      oprot.writeFieldEnd()
    if self.port is not None:
      oprot.writeFieldBegin('port', TType.I16, 2)
      oprot.writeI16(self.port)
      oprot.writeFieldEnd()
    if self.service_name is not None:
      oprot.writeFieldBegin('service_name', TType.STRING, 3)
      oprot.writeString(self.service_name)
      oprot.writeFieldEnd()
    if self.ipv6 is not None:
      oprot.writeFieldBegin('ipv6', TType.STRING, 4)
      oprot.writeString(self.ipv6)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.ipv4)
    value = (value * 31) ^ hash(self.port)
    value = (value * 31) ^ hash(self.service_name)
    value = (value * 31) ^ hash(self.ipv6)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in six.iteritems(self.__dict__)]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class Annotation(object):
  """
  An annotation is similar to a log statement. It includes a host field which
  allows these events to be attributed properly, and also aggregatable.

  Attributes:
   - timestamp: Microseconds from epoch.

  This value should use the most precise value possible. For example,
  gettimeofday or syncing nanoTime against a tick of currentTimeMillis.
   - value
   - host: Always the host that recorded the event. By specifying the host you allow
  rollup of all events (such as client requests to a service) by IP address.
  """

  thrift_spec = (
    None, # 0
    (1, TType.I64, 'timestamp', None, None, ), # 1
    (2, TType.STRING, 'value', None, None, ), # 2
    (3, TType.STRUCT, 'host', (Endpoint, Endpoint.thrift_spec), None, ), # 3
  )

  def __init__(self, timestamp=None, value=None, host=None,):
    self.timestamp = timestamp
    self.value = value
    self.host = host

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.I64:
          self.timestamp = iprot.readI64()
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.STRING:
          self.value = iprot.readString()
        else:
          iprot.skip(ftype)
      elif fid == 3:
        if ftype == TType.STRUCT:
          self.host = Endpoint()
          self.host.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('Annotation')
    if self.timestamp is not None:
      oprot.writeFieldBegin('timestamp', TType.I64, 1)
      oprot.writeI64(self.timestamp)
      oprot.writeFieldEnd()
    if self.value is not None:
      oprot.writeFieldBegin('value', TType.STRING, 2)
      oprot.writeString(self.value)
      oprot.writeFieldEnd()
    if self.host is not None:
      oprot.writeFieldBegin('host', TType.STRUCT, 3)
      self.host.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.timestamp)
    value = (value * 31) ^ hash(self.value)
    value = (value * 31) ^ hash(self.host)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in six.iteritems(self.__dict__)]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class BinaryAnnotation(object):
  """
  Binary annotations are tags applied to a Span to give it context. For
  example, a binary annotation of "http.uri" could the path to a resource in a
  RPC call.

  Binary annotations of type STRING are always queryable, though more a
  historical implementation detail than a structural concern.

  Binary annotations can repeat, and vary on the host. Similar to Annotation,
  the host indicates who logged the event. This allows you to tell the
  difference between the client and server side of the same key. For example,
  the key "http.uri" might be different on the client and server side due to
  rewriting, like "/api/v1/myresource" vs "/myresource. Via the host field,
  you can see the different points of view, which often help in debugging.

  Attributes:
   - key
   - value
   - annotation_type
   - host: The host that recorded tag, which allows you to differentiate between
  multiple tags with the same key. There are two exceptions to this.

  When the key is CLIENT_ADDR or SERVER_ADDR, host indicates the source or
  destination of an RPC. This exception allows zipkin to display network
  context of uninstrumented services, or clients such as web browsers.
  """

  thrift_spec = (
    None, # 0
    (1, TType.STRING, 'key', None, None, ), # 1
    (2, TType.STRING, 'value', None, None, ), # 2
    (3, TType.I32, 'annotation_type', None, None, ), # 3
    (4, TType.STRUCT, 'host', (Endpoint, Endpoint.thrift_spec), None, ), # 4
  )

  def __init__(self, key=None, value=None, annotation_type=None, host=None,):
    self.key = key
    self.value = value
    self.annotation_type = annotation_type
    self.host = host

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRING:
          self.key = iprot.readString()
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.STRING:
          self.value = iprot.readString()
        else:
          iprot.skip(ftype)
      elif fid == 3:
        if ftype == TType.I32:
          self.annotation_type = iprot.readI32()
        else:
          iprot.skip(ftype)
      elif fid == 4:
        if ftype == TType.STRUCT:
          self.host = Endpoint()
          self.host.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('BinaryAnnotation')
    if self.key is not None:
      oprot.writeFieldBegin('key', TType.STRING, 1)
      oprot.writeString(self.key)
      oprot.writeFieldEnd()
    if self.value is not None:
      oprot.writeFieldBegin('value', TType.STRING, 2)
      oprot.writeString(self.value)
      oprot.writeFieldEnd()
    if self.annotation_type is not None:
      oprot.writeFieldBegin('annotation_type', TType.I32, 3)
      oprot.writeI32(self.annotation_type)
      oprot.writeFieldEnd()
    if self.host is not None:
      oprot.writeFieldBegin('host', TType.STRUCT, 4)
      self.host.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.key)
    value = (value * 31) ^ hash(self.value)
    value = (value * 31) ^ hash(self.annotation_type)
    value = (value * 31) ^ hash(self.host)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in six.iteritems(self.__dict__)]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class Span(object):
  """
  A trace is a series of spans (often RPC calls) which form a latency tree.

  The root span is where trace_id = id and parent_id = Nil. The root span is
  usually the longest interval in the trace, starting with a SERVER_RECV
  annotation and ending with a SERVER_SEND.

  Attributes:
   - trace_id
   - name: Span name in lowercase, rpc method for example

  Conventionally, when the span name isn't known, name = "unknown".
   - id
   - parent_id
   - annotations
   - binary_annotations
   - debug
   - timestamp: Microseconds from epoch of the creation of this span.

  This value should be set directly by instrumentation, using the most
  precise value possible. For example, gettimeofday or syncing nanoTime
  against a tick of currentTimeMillis.

  For compatibilty with instrumentation that precede this field, collectors
  or span stores can derive this via Annotation.timestamp.
  For example, SERVER_RECV.timestamp or CLIENT_SEND.timestamp.

  This field is optional for compatibility with old data: first-party span
  stores are expected to support this at time of introduction.
   - duration: Measurement of duration in microseconds, used to support queries.

  This value should be set directly, where possible. Doing so encourages
  precise measurement decoupled from problems of clocks, such as skew or NTP
  updates causing time to move backwards.

  For compatibilty with instrumentation that precede this field, collectors
  or span stores can derive this by subtracting Annotation.timestamp.
  For example, SERVER_SEND.timestamp - SERVER_RECV.timestamp.

  If this field is persisted as unset, zipkin will continue to work, except
  duration query support will be implementation-specific. Similarly, setting
  this field non-atomically is implementation-specific.

  This field is i64 vs i32 to support spans longer than 35 minutes.
   - trace_id_high: Optional unique 8-byte additional identifier for a trace. If non zero, this
  means the trace uses 128 bit traceIds instead of 64 bit.
  """

  thrift_spec = (
    None, # 0
    (1, TType.I64, 'trace_id', None, None, ), # 1
    None, # 2
    (3, TType.STRING, 'name', None, None, ), # 3
    (4, TType.I64, 'id', None, None, ), # 4
    (5, TType.I64, 'parent_id', None, None, ), # 5
    (6, TType.LIST, 'annotations', (TType.STRUCT,(Annotation, Annotation.thrift_spec)), None, ), # 6
    None, # 7
    (8, TType.LIST, 'binary_annotations', (TType.STRUCT,(BinaryAnnotation, BinaryAnnotation.thrift_spec)), None, ), # 8
    (9, TType.BOOL, 'debug', None, False, ), # 9
    (10, TType.I64, 'timestamp', None, None, ), # 10
    (11, TType.I64, 'duration', None, None, ), # 11
    (12, TType.I64, 'trace_id_high', None, None, ), # 12
  )

  def __init__(self, trace_id=None, name=None, id=None, parent_id=None, annotations=None, binary_annotations=None, debug=thrift_spec[9][4], timestamp=None, duration=None, trace_id_high=None,):
    self.trace_id = trace_id
    self.name = name
    self.id = id
    self.parent_id = parent_id
    self.annotations = annotations
    self.binary_annotations = binary_annotations
    self.debug = debug
    self.timestamp = timestamp
    self.duration = duration
    self.trace_id_high = trace_id_high

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.I64:
          self.trace_id = iprot.readI64()
        else:
          iprot.skip(ftype)
      elif fid == 3:
        if ftype == TType.STRING:
          self.name = iprot.readString()
        else:
          iprot.skip(ftype)
      elif fid == 4:
        if ftype == TType.I64:
          self.id = iprot.readI64()
        else:
          iprot.skip(ftype)
      elif fid == 5:
        if ftype == TType.I64:
          self.parent_id = iprot.readI64()
        else:
          iprot.skip(ftype)
      elif fid == 6:
        if ftype == TType.LIST:
          self.annotations = []
          (_etype3, _size0) = iprot.readListBegin()
          for _i4 in xrange(_size0):
            _elem5 = Annotation()
            _elem5.read(iprot)
            self.annotations.append(_elem5)
          iprot.readListEnd()
        else:
          iprot.skip(ftype)
      elif fid == 8:
        if ftype == TType.LIST:
          self.binary_annotations = []
          (_etype9, _size6) = iprot.readListBegin()
          for _i10 in xrange(_size6):
            _elem11 = BinaryAnnotation()
            _elem11.read(iprot)
            self.binary_annotations.append(_elem11)
          iprot.readListEnd()
        else:
          iprot.skip(ftype)
      elif fid == 9:
        if ftype == TType.BOOL:
          self.debug = iprot.readBool()
        else:
          iprot.skip(ftype)
      elif fid == 10:
        if ftype == TType.I64:
          self.timestamp = iprot.readI64()
        else:
          iprot.skip(ftype)
      elif fid == 11:
        if ftype == TType.I64:
          self.duration = iprot.readI64()
        else:
          iprot.skip(ftype)
      elif fid == 12:
        if ftype == TType.I64:
          self.trace_id_high = iprot.readI64()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('Span')
    if self.trace_id is not None:
      oprot.writeFieldBegin('trace_id', TType.I64, 1)
      oprot.writeI64(self.trace_id)
      oprot.writeFieldEnd()
    if self.name is not None:
      oprot.writeFieldBegin('name', TType.STRING, 3)
      oprot.writeString(self.name)
      oprot.writeFieldEnd()
    if self.id is not None:
      oprot.writeFieldBegin('id', TType.I64, 4)
      oprot.writeI64(self.id)
      oprot.writeFieldEnd()
    if self.parent_id is not None:
      oprot.writeFieldBegin('parent_id', TType.I64, 5)
      oprot.writeI64(self.parent_id)
      oprot.writeFieldEnd()
    if self.annotations is not None:
      oprot.writeFieldBegin('annotations', TType.LIST, 6)
      oprot.writeListBegin(TType.STRUCT, len(self.annotations))
      for iter12 in self.annotations:
        iter12.write(oprot)
      oprot.writeListEnd()
      oprot.writeFieldEnd()
    if self.binary_annotations is not None:
      oprot.writeFieldBegin('binary_annotations', TType.LIST, 8)
      oprot.writeListBegin(TType.STRUCT, len(self.binary_annotations))
      for iter13 in self.binary_annotations:
        iter13.write(oprot)
      oprot.writeListEnd()
      oprot.writeFieldEnd()
    if self.debug is not None:
      oprot.writeFieldBegin('debug', TType.BOOL, 9)
      oprot.writeBool(self.debug)
      oprot.writeFieldEnd()
    if self.timestamp is not None:
      oprot.writeFieldBegin('timestamp', TType.I64, 10)
      oprot.writeI64(self.timestamp)
      oprot.writeFieldEnd()
    if self.duration is not None:
      oprot.writeFieldBegin('duration', TType.I64, 11)
      oprot.writeI64(self.duration)
      oprot.writeFieldEnd()
    if self.trace_id_high is not None:
      oprot.writeFieldBegin('trace_id_high', TType.I64, 12)
      oprot.writeI64(self.trace_id_high)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.trace_id)
    value = (value * 31) ^ hash(self.name)
    value = (value * 31) ^ hash(self.id)
    value = (value * 31) ^ hash(self.parent_id)
    value = (value * 31) ^ hash(self.annotations)
    value = (value * 31) ^ hash(self.binary_annotations)
    value = (value * 31) ^ hash(self.debug)
    value = (value * 31) ^ hash(self.timestamp)
    value = (value * 31) ^ hash(self.duration)
    value = (value * 31) ^ hash(self.trace_id_high)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in six.iteritems(self.__dict__)]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class Response(object):
  """
  Attributes:
   - ok
  """

  thrift_spec = (
    None, # 0
    (1, TType.BOOL, 'ok', None, None, ), # 1
  )

  def __init__(self, ok=None,):
    self.ok = ok

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.BOOL:
          self.ok = iprot.readBool()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('Response')
    if self.ok is not None:
      oprot.writeFieldBegin('ok', TType.BOOL, 1)
      oprot.writeBool(self.ok)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    if self.ok is None:
      raise TProtocol.TProtocolException(message='Required field ok is unset!')
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.ok)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in six.iteritems(self.__dict__)]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)



# Modified by SignalFx
# Copyright (c) 2016 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import threading
import time


class PeriodicCallback(object):
    """
    Thread-based implementation of tornado.io_loop.PeriodicCallback:
    Schedules the given callback to be called periodically in a background thread.

    The callback is called every ``callback_time`` milliseconds.

    If the callback runs for longer than ``callback_time`` milliseconds,
    subsequent invocations will be skipped.

    `start` must be called after the `PeriodicCallback` is created.
    """

    def __init__(self, callback, callback_time):
        self.callback = callback
        if callback_time < 0:
            raise ValueError('callback_time must be positive')
        self.callback_time = callback_time
        self._scheduler_sleep = callback_time / 1000.0
        self._running = False
        self._lock = threading.Lock()
        self._scheduler_thread = None
        self._callback_thread = None

    def start(self):
        with self._lock:
            if self._running:
                return
            self._running = True
            if self._scheduler_thread is None:
                self._scheduler_thread = threading.Thread(target=self._scheduler)
                self._scheduler_thread.daemon = True
                self._scheduler_thread.start()

    def stop(self):
        with self._lock:
            if self._running:
                self._running = False

            if self._scheduler_thread is not None:
                self._scheduler_thread.join()
                self._scheduler_thread = None
            if self._callback_thread is not None:
                self._callback_thread.join()
                self._callback_thread = None

    def _scheduler(self):
        while self._running:
            if self._callback_thread:
                if not self._callback_thread.is_alive():
                    self._callback_thread.join()
                    self._callback_thread = None
            if self._callback_thread is None:
                self._callback_thread = threading.Thread(target=self.callback)
                self._callback_thread.daemon = True
                self._callback_thread.start()

            time.sleep(self._scheduler_sleep)



# Modified by SignalFx
# Copyright (c) 2016 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time

import mock
import pytest
from jaeger_client import ConstSampler, Tracer


@pytest.fixture(scope='function')
def tracer():
    reporter = mock.MagicMock()
    sampler = ConstSampler(True)
    return Tracer(
        service_name='test_service_1', reporter=reporter, sampler=sampler)


def wait_for(fn, ms=1000):
    """
    Wait until fn() returns truth, but not longer than specified duration
    (1000ms by default).  Returns True if fn() ever truthy, otherwise False.
    """
    for i in range(ms):
        if fn():
            return True
        time.sleep(0.001)
    return False



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import absolute_import

import opentracing
from opentracing import Tracer, Format


def test_new_trace():
    tracer = Tracer()

    span = tracer.start_span(operation_name='test')
    span.set_baggage_item('Fry', 'Leela')
    span.set_tag('x', 'y')
    span.log_event('z')

    child = tracer.start_span(operation_name='child',
                              references=opentracing.child_of(span.context))
    child.log_event('w')
    assert child.get_baggage_item('Fry') is None
    carrier = {}
    tracer.inject(
        span_context=child.context,
        format=Format.TEXT_MAP,
        carrier=carrier)
    assert carrier == dict()
    child.finish()

    span.finish()


def test_join_trace():
    tracer = Tracer()

    span_ctx = tracer.extract(format=Format.TEXT_MAP, carrier={})
    span = tracer.start_span(operation_name='test',
                             references=opentracing.child_of(span_ctx))
    span.set_tag('x', 'y')
    span.set_baggage_item('a', 'b')
    span.log_event('z')

    child = tracer.start_span(operation_name='child',
                              references=opentracing.child_of(span.context))
    child.log_event('w')
    child.finish()

    span.finish()



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import logging

from crossdock.thrift_gen.tracetest.ttypes import JoinTraceRequest, StartTraceRequest, \
    Downstream, Transport, TraceResponse, ObservedSpan


#
# Serializers for the downstream calls
#
def set_downstream_object_values(downstream_object, json_obj):
    return set_traced_service_object_values(downstream_object, json_obj, downstream_from_struct)


def join_trace_request_from_json(json_request):
    request = JoinTraceRequest()
    set_downstream_object_values(request, json.loads(json_request))
    return request


def start_trace_request_from_json(json_request):
    request = StartTraceRequest()
    set_downstream_object_values(request, json.loads(json_request))
    return request


def downstream_from_struct(json_obj):
    downstream = Downstream()
    set_downstream_object_values(downstream, json_obj)
    return downstream


def join_trace_request_to_json(downstream, server_role):
    req = {}
    if downstream is not None:
        req['downstream'] = traced_service_object_to_json(downstream)
    if server_role is not None:
        req['serverRole'] = str(server_role)
    return json.dumps(req)

#
# Serializers for the upstream responses
#


def set_upstream_object_values(obj, json_obj):
    return set_traced_service_object_values(obj, json_obj, traceresponse_from_struct)


def observed_span_from_struct(json_obj):
    os = ObservedSpan()
    set_upstream_object_values(os, json_obj)
    return os


def traceresponse_from_struct(json_obj):
    tr = TraceResponse()
    set_upstream_object_values(tr, json_obj)
    return tr


def traceresponse_from_json(json_str):
    try:
        return traceresponse_from_struct(json.loads(json_str))
    except:  # noqa: E722
        logging.exception('Failed to parse JSON')
        raise

# Generic


def class_keys(obj):
    return [a for a in dir(obj) if not a.startswith('__') and not
            callable(getattr(obj, a)) and not
            a == 'type_spec']


def traced_service_object_to_json(obj):
    json_response = {}
    for k in class_keys(obj):
        if k == 'downstream':
            if obj.downstream is not None:
                json_response['downstream'] = traced_service_object_to_json(obj.downstream)
        elif k == 'transport':
            if obj.transport is not None:
                json_response['transport'] = Transport._VALUES_TO_NAMES[obj.transport]
        elif k == 'span':
            if obj.span is not None:
                json_response['span'] = traced_service_object_to_json(obj.span)
        else:
            json_response[k] = getattr(obj, k)
    return json_response


def set_traced_service_object_values(obj, values, downstream_func):
    for k in values.iterkeys():
        if hasattr(obj, k):
            if k == 'downstream':
                if values[k] is not None:
                    obj.downstream = downstream_func(values[k])
            elif k == 'transport':
                if values[k] is not None:
                    obj.transport = Transport._NAMES_TO_VALUES[values[k]]
            elif k == 'span':
                if values[k] is not None:
                    obj.span = observed_span_from_struct(values[k])
            else:
                setattr(obj, k, values[k])



# Modified by SignalFx
# Copyright (c) 2016 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import random
import six
import socket

import pytest

from opentracing import Format, child_of
from opentracing.ext import tags as ext_tags
from jaeger_client import ConstSampler, SpanContext, Tracer
from jaeger_client import constants as c
from jaeger_client.codecs import B3Codec


def find_tag(span, key, tag_type='str'):
    for tag in span.tags:
        if tag.key == key:
            if tag_type == 'str':
                return tag.vStr
            elif tag_type == 'bool':
                return tag.vBool
    return None


def test_start_trace(tracer):
    assert type(tracer) is Tracer
    with mock.patch.object(random.Random, 'getrandbits') as mock_random, \
            mock.patch('time.time') as mock_timestamp:
        mock_random.return_value = 12345
        mock_timestamp.return_value = 54321

        span = tracer.start_span('test')
        span.set_tag(ext_tags.SPAN_KIND, ext_tags.SPAN_KIND_RPC_SERVER)
        assert span, 'Span must not be nil'
        assert span.tracer == tracer, 'Tracer must be referenced from span'
        assert find_tag(span, 'span.kind') == ext_tags.SPAN_KIND_RPC_SERVER, \
            'Span must be server-side'
        assert span.trace_id == 12345, 'Must match trace_id'
        assert span.is_sampled(), 'Must be sampled'
        assert span.parent_id is None, 'Must not have parent_id (root span)'
        assert span.start_time == 54321, 'Must match timestamp'

        span.finish()
        assert span.end_time is not None, 'Must have end_time defined'
        tracer.reporter.assert_called_once()

    tracer.close()


def test_forced_sampling(tracer):
    tracer.sampler = ConstSampler(False)
    span = tracer.start_span('test2',
                             tags={ext_tags.SAMPLING_PRIORITY: 1})
    assert span.is_sampled()
    assert span.is_debug()


@pytest.mark.parametrize('mode,', ['arg', 'ref'])
def test_start_child(tracer, mode):
    root = tracer.start_span('test')
    if mode == 'arg':
        span = tracer.start_span('test', child_of=root.context)
    elif mode == 'ref':
        span = tracer.start_span('test', references=child_of(root.context))
    else:
        raise ValueError('bad mode')
    span.set_tag(ext_tags.SPAN_KIND, ext_tags.SPAN_KIND_RPC_SERVER)
    assert span.is_sampled(), 'Must be sampled'
    assert span.trace_id == root.trace_id, 'Must have the same trace id'
    assert span.parent_id == root.span_id, 'Must inherit parent id'
    span.finish()
    assert span.end_time is not None, 'Must have end_time set'
    tracer.reporter.assert_called_once()
    tracer.close()


@pytest.mark.parametrize('one_span_per_rpc,', [True, False])
def test_one_span_per_rpc(tracer, one_span_per_rpc):
    tracer.one_span_per_rpc = one_span_per_rpc
    span = tracer.start_span('client')
    span.set_tag(ext_tags.SPAN_KIND, ext_tags.SPAN_KIND_RPC_CLIENT)
    child = tracer.start_span(
        'server',
        references=child_of(span.context),
        tags={ext_tags.SPAN_KIND: ext_tags.SPAN_KIND_RPC_SERVER},
    )
    assert span.trace_id == child.trace_id, 'Must have the same trace ids'
    if one_span_per_rpc:
        assert span.span_id == child.span_id, 'Must have the same span ids'
    else:
        assert span.span_id != child.span_id, 'Must have different span ids'


def test_child_span(tracer):
    span = tracer.start_span('test')
    child = tracer.start_span('child', references=child_of(span.context))
    child.set_tag(ext_tags.SPAN_KIND, ext_tags.SPAN_KIND_RPC_CLIENT)
    child.set_tag('bender', 'is great')
    child.log_event('kiss-my-shiny-metal-...')
    child.finish()
    span.finish()
    tracer.reporter.report_span.assert_called_once()
    assert len(span.logs) == 0, 'Parent span is Local, must not have events'
    assert len(child.logs) == 1, 'Child must have one events'

    tracer.sampler = ConstSampler(False)
    span = tracer.start_span('test')
    child = tracer.start_span('child', references=child_of(span.context))
    child.set_tag('bender', 'is great')
    child.log_event('kiss-my-shiny-metal-...')
    child.finish()
    span.finish()
    assert len(child.logs) == 0, 'Must have no events, not sampled'
    assert len(child.tags) == 0, 'Must have no attributes, not sampled'
    tracer.close()


def test_sampler_effects(tracer):
    tracer.sampler = ConstSampler(True)
    span = tracer.start_span('test')
    assert span.is_sampled(), 'Must be sampled'

    tracer.sampler = ConstSampler(False)
    span = tracer.start_span('test')
    assert not span.is_sampled(), 'Must not be sampled'
    tracer.close()


@pytest.mark.parametrize('inject_mode', ['span', 'context'])
def test_serialization(tracer, inject_mode):
    span = tracer.start_span('help')
    carrier = {}
    if inject_mode == 'span':
        injectable = span
    elif inject_mode == 'context':
        injectable = span.context
    else:
        raise ValueError('bad inject_mode')
    tracer.inject(
        span_context=injectable, format=Format.TEXT_MAP, carrier=carrier
    )
    assert len(carrier) > 0
    assert 'uber-trace-id' in carrier
    ctx = tracer.extract(Format.TEXT_MAP, carrier)
    assert ctx is not None
    assert ctx.trace_id == span.trace_id
    assert ctx.span_id == span.span_id
    assert ctx.parent_id == span.parent_id
    assert ctx.flags == span.flags


def test_serialization_error(tracer):
    span = 'span'
    carrier = {}
    with pytest.raises(ValueError):
        tracer.inject(
            span_context=span, format=Format.TEXT_MAP, carrier=carrier
        )


def test_tracer_tags():
    reporter = mock.MagicMock()
    sampler = ConstSampler(True)

    with mock.patch('socket.gethostname', return_value='dream-host.com'):
        t = Tracer(service_name='x', reporter=reporter, sampler=sampler)
        assert t.tags.get('hostname') == 'dream-host.com'
        assert 'ip' in t.tags
        assert 'jaeger.version' in t.tags


def test_tracer_tags_passed_to_reporter():
    reporter = mock.MagicMock()
    reporter.set_process = mock.MagicMock()
    sampler = ConstSampler(True)
    tracer = Tracer(
        service_name='x', reporter=reporter, sampler=sampler,
        max_tag_value_length=123,
    )
    reporter.set_process.assert_called_once_with(
        service_name='x', tags=tracer.tags, max_length=123,
    )


def test_tracer_tags_no_hostname():
    reporter = mock.MagicMock()
    sampler = ConstSampler(True)

    from jaeger_client.tracer import logger
    with mock.patch.object(logger, 'exception') as mock_log:
        with mock.patch('socket.gethostname',
                        side_effect=['host', socket.timeout()]):
            Tracer(service_name='x', reporter=reporter, sampler=sampler)
        assert mock_log.call_count == 1


@pytest.mark.parametrize('span_type,expected_tags', [
    ('root', {
        'sampler.type': 'const',
        'sampler.param': True,
    }),
    ('child', {
        'sampler.type': None,
        'sampler.param': None,
    }),
    ('rpc-server', {
        'sampler.type': None,
        'sampler.param': None,
    }),
])
def test_tracer_tags_on_root_span(span_type, expected_tags):
    reporter = mock.MagicMock()
    sampler = ConstSampler(True)
    with mock.patch('socket.gethostname', return_value='dream-host.com'):
        tracer = Tracer(service_name='x',
                        reporter=reporter,
                        sampler=sampler,
                        tags={'global-tag': 'global-tag'})
        span = tracer.start_span(operation_name='root')
        if span_type == 'child':
            span = tracer.start_span('child', child_of=span)
        if span_type == 'rpc-server':
            span = tracer.start_span(
                'child', child_of=span.context,
                tags={ext_tags.SPAN_KIND: ext_tags.SPAN_KIND_RPC_SERVER}
            )
        for key, value in six.iteritems(expected_tags):
            found_tag = find_tag(span, key, type(value).__name__)
            if value is None:
                assert found_tag is None, 'test (%s)' % span_type
                continue

            assert found_tag == value, \
                'test (%s): expecting tag %s=%s' % (span_type, key, value)


def test_tracer_local_root_span_tags():
    reporter = mock.MagicMock()
    sampler = ConstSampler(True)

    tracer = Tracer(service_name='x',
                    reporter=reporter,
                    sampler=sampler,
                    root_span_tags={'root-tag1': 'value1'})


    root = tracer.start_span('root')
    child = tracer.start_span('child', child_of=root)
    child.finish()
    root.finish()

    tag = find_tag(root, "root-tag1")
    assert tag is not None
    assert tag == "value1"

    tag = find_tag(child, "root-tag1")
    assert tag is None


    tracer.reporter.report_span.assert_called_once()

    ctx = B3Codec().extract({
        "X-B3-Sampled": "1",
        "X-B3-Spanid": "60e6deb669a8497a",
        "X-B3-Traceid": "9e64f559d130e123",
        "X-B3-Parentspanid": "42b6fe7d9f34b1e8"
    })
    assert ctx is not None
    rpc_span = tracer.start_span('child', child_of=ctx)
    rpc_span.finish()

    tag = find_tag(rpc_span, "root-tag1")
    assert tag is not None
    assert tag == "value1"


def test_tracer_override_codecs():
    reporter = mock.MagicMock()
    sampler = ConstSampler(True)
    codecs = {
        'extra_codec': 'codec_placeholder',
        Format.BINARY: 'overridden_binary_codec'

    }
    with mock.patch('socket.gethostname', return_value='dream-host.com'):
        tracer = Tracer(service_name='x', reporter=reporter, sampler=sampler,
                        extra_codecs=codecs)
        assert tracer.codecs['extra_codec'] == 'codec_placeholder',\
                                               'Extra codec not found'
        assert tracer.codecs[Format.BINARY] == 'overridden_binary_codec',\
                                               'Binary format codec not overridden'


def test_tracer_hostname_tag():
    reporter = mock.MagicMock()
    sampler = ConstSampler(True)
    tracer = Tracer(
        service_name='x',
        tags={c.JAEGER_HOSTNAME_TAG_KEY: 'jaeger-client-app.local'},
        reporter=reporter,
        sampler=sampler,
    )

    assert tracer.tags[c.JAEGER_HOSTNAME_TAG_KEY] == 'jaeger-client-app.local'


def test_tracer_ip_tag():
    reporter = mock.MagicMock()
    sampler = ConstSampler(True)
    tracer = Tracer(
        service_name='x',
        tags={c.JAEGER_IP_TAG_KEY: '192.0.2.3'},
        reporter=reporter,
        sampler=sampler,
    )

    assert tracer.tags[c.JAEGER_IP_TAG_KEY] == '192.0.2.3'


def test_tracer_throttler():
    tracer = Tracer(
        service_name='x',
        reporter=mock.MagicMock(),
        sampler=mock.MagicMock(),
        throttler=mock.MagicMock())
    tracer.throttler.is_allowed.return_value = True
    assert tracer.is_debug_allowed()
    tracer.throttler.is_allowed.return_value = False
    assert not tracer.is_debug_allowed()
    debug_span_context = SpanContext.with_debug_id('debug-id')
    span = tracer.start_span('test-operation', child_of=debug_span_context)
    assert not span.is_debug()


def test_tracer_128bit_trace_id():
    reporter = mock.MagicMock()
    sampler = mock.MagicMock()
    tracer = Tracer(
        service_name='x',
        reporter=reporter,
        sampler=sampler,
    )
    assert tracer.max_trace_id_bits == c._max_id_bits

    tracer = Tracer(
        service_name='x',
        reporter=reporter,
        sampler=sampler,
        generate_128bit_trace_id=True,
    )
    assert tracer.max_trace_id_bits == c._max_trace_id_bits



# Modified by SignalFx
# Copyright (c) 2016-2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
import os
import unittest
import opentracing.tracer
from jaeger_client import Config, ConstSampler, ProbabilisticSampler, RateLimitingSampler
from jaeger_client.config import DEFAULT_THROTTLER_PORT
from jaeger_client.metrics import MetricsFactory
from jaeger_client.reporter import NullReporter
from jaeger_client.senders import HTTPSender
from jaeger_client import constants


class ConfigTests(unittest.TestCase):

    def setUp(self):
        if Config.initialized():
            Config._initialized = False

    def test_enabled(self):
        c = Config({'enabled': True}, service_name='x')
        assert c.enabled
        c = Config({'enabled': False}, service_name='x')
        assert not c.enabled

    def test_reporter_batch_size(self):
        c = Config({'reporter_batch_size': 12345}, service_name='x')
        assert c.reporter_batch_size == 12345
        c = Config({}, service_name='x')
        assert c.reporter_batch_size == 10

    def test_tags(self):
        os.environ['JAEGER_TAGS'] = 'a=b,c=d'
        c = Config({'tags': {'e': 'f'}}, service_name='x')
        assert c.tags == {'a': 'b', 'c': 'd', 'e': 'f'}
        c.create_tracer(NullReporter(), ConstSampler(True))

    def test_no_sampler(self):
        c = Config({}, service_name='x')
        assert c.sampler is None

    def test_const_sampler(self):
        c = Config({'sampler': {'type': 'const', 'param': True}},
                   service_name='x')
        assert type(c.sampler) is ConstSampler
        assert c.sampler.decision
        c = Config({'sampler': {'type': 'const', 'param': False}},
                   service_name='x')
        assert type(c.sampler) is ConstSampler
        assert not c.sampler.decision

    def test_probabilistic_sampler(self):
        with self.assertRaises(Exception):
            cfg = {'sampler': {'type': 'probabilistic', 'param': 'xx'}}
            Config(cfg, service_name='x').sampler
        c = Config({'sampler': {'type': 'probabilistic', 'param': 0.5}},
                   service_name='x')
        assert type(c.sampler) is ProbabilisticSampler
        assert c.sampler.rate == 0.5

    def test_rate_limiting_sampler(self):
        with self.assertRaises(Exception):
            cfg = {'sampler': {'type': 'rate_limiting', 'param': 'xx'}}
            Config(cfg, service_name='x').sampler
        c = Config({'sampler': {'type': 'rate_limiting', 'param': 1234}},
                   service_name='x')
        assert type(c.sampler) is RateLimitingSampler
        assert c.sampler.traces_per_second == 1234

    def test_bad_sampler(self):
        c = Config({'sampler': {'type': 'bad-sampler'}}, service_name='x')
        with self.assertRaises(ValueError):
            c.sampler.is_sampled(0)

    def test_agent_reporting_host(self):
        c = Config({}, service_name='x')
        assert c.local_agent_reporting_host == 'localhost'

        c = Config({'local_agent': {'reporting_host': 'jaeger.local'}}, service_name='x')
        assert c.local_agent_reporting_host == 'jaeger.local'

        os.environ['JAEGER_AGENT_HOST'] = 'jaeger-env.local'
        c = Config({}, service_name='x')
        assert c.local_agent_reporting_host == 'jaeger-env.local'

    def test_max_tag_value_length(self):
        c = Config({}, service_name='x')
        assert c.max_tag_value_length == constants.MAX_TAG_VALUE_LENGTH

        c = Config({'max_tag_value_length': 333}, service_name='x')
        assert c.max_tag_value_length == 333

        t = c.create_tracer(NullReporter(), ConstSampler(True))
        assert t.max_tag_value_length == 333

    def test_propagation(self):
        c = Config({}, service_name='x')
        assert c.propagation == {}

        c = Config({'propagation': 'b3'}, service_name='x')
        assert len(c.propagation) == 1

    def test_throttler(self):
        c = Config({
            'throttler': {}
        }, service_name='x')
        assert not c.throttler_group()
        assert c.throttler_port == DEFAULT_THROTTLER_PORT
        assert c.throttler_refresh_interval == constants.DEFAULT_THROTTLER_REFRESH_INTERVAL

        c = Config({
            'throttler': {
                'port': '1234',
                'refresh_interval': '10'
            }
        }, service_name='x')
        assert c.throttler_group()
        assert c.throttler_port == 1234
        assert c.throttler_refresh_interval == 10

        c = Config({}, service_name='x')
        assert c.throttler_group() is None
        assert c.throttler_port is None
        assert c.throttler_refresh_interval is None

    def test_for_unexpected_config_entries(self):
        with self.assertRaises(Exception):
            Config({'unexpected': 'value'}, validate=True)

    def test_reporter_queue_size_valid(self):
        config = Config({'reporter_queue_size': 100}, service_name='x', validate=True)
        assert config.reporter_queue_size == 100

    def test_missing_service_name(self):
        with self.assertRaises(ValueError):
            Config({})

    def test_disable_metrics(self):
        config = Config({'metrics': False}, service_name='x')
        assert isinstance(config._metrics_factory, MetricsFactory)

    def test_initialize_tracer(self):
        c = Config({}, service_name='x')
        tracer = c.initialize_tracer()

        assert opentracing.tracer == tracer
        tracer.close()

    def test_default_local_agent_reporting_port(self):
        c = Config({}, service_name='x')
        assert c.local_agent_reporting_port == 6831

    def test_generate_128bit_trace_id(self):
        c = Config({}, service_name='x')
        assert c.generate_128bit_trace_id is False

        c = Config({'generate_128bit_trace_id': True}, service_name='x')
        assert c.generate_128bit_trace_id is True

        os.environ['JAEGER_TRACEID_128BIT'] = 'true'
        c = Config({'generate_128bit_trace_id': False}, service_name='x')
        assert c.generate_128bit_trace_id is False

        c = Config({}, service_name='x')
        assert c.generate_128bit_trace_id is True

        os.environ.pop('JAEGER_TRACEID_128BIT')
        assert os.getenv('JAEGER_TRACEID_128BIT', None) is None

    def test_global_tracer_initializaion(self):
        c = Config({}, service_name='x')
        tracer = c.initialize_tracer()
        assert tracer
        attempt = c.initialize_tracer()
        assert attempt is None
        tracer.close()

    def test_jaeger_endpoint(self):
        c = Config({'jaeger_endpoint': 'some_endpoint'}, service_name='x')
        assert c.jaeger_endpoint == 'some_endpoint'
        os.environ['JAEGER_ENDPOINT'] = 'SomeEndpoint'
        c = Config({}, service_name='x')
        assert c.jaeger_endpoint == 'SomeEndpoint'
        del os.environ['JAEGER_ENDPOINT']

    def test_jaeger_auth_token(self):
        c = Config({'jaeger_auth_token': 'some_token'},
                   service_name='x', validate=True)
        assert c.jaeger_auth_token == 'some_token'
        os.environ['JAEGER_AUTH_TOKEN'] = 'SomeToken'
        c = Config({}, service_name='x')
        assert c.jaeger_auth_token == 'SomeToken'
        del os.environ['JAEGER_AUTH_TOKEN']

    def test_jaeger_user_and_password(self):
        c = Config({'jaeger_user': 'some_user',
                    'jaeger_password': 'some_password'},
                   service_name='x', validate=True)
        assert c.jaeger_user == 'some_user'
        assert c.jaeger_password == 'some_password'
        os.environ['JAEGER_USER'] = 'SomeUser'
        os.environ['JAEGER_PASSWORD'] = 'SomePassword'
        c = Config({}, service_name='x')
        assert c.jaeger_user == 'SomeUser'
        assert c.jaeger_password == 'SomePassword'
        del os.environ['JAEGER_USER']
        del os.environ['JAEGER_PASSWORD']

    def test_jaeger_auth_token_and_basic_mutually_exclusive(self):
        with self.assertRaises(ValueError) as e:
            Config({'jaeger_auth_token': 'some_token',
                    'jaeger_user': 'some_user',
                    'jaeger_password': 'some_password'},
                   service_name='x', validate=True)
        assert e.exception.args[0] == ('Cannot accept both jaeger_auth_token and '
                                       'jaeger_user/jaeger_password for authentication')

    def test_jaeger_user_and_password_required(self):
        for cfg in ({'jaeger_user': 'some_user'},
                    {'jaeger_password': 'some_password'}):
            with self.assertRaises(ValueError) as e:
                Config(cfg, service_name='x', validate=True)

            assert e.exception.args[0] == ('Must provide both jaeger_user and '
                                           'jaeger_password for authentication.')

    def test_specifying_jaeger_endpoint_creates_http_sender(self):
        c = Config({'jaeger_endpoint': 'some_endpoint'}, service_name='x')
        tracer = c.initialize_tracer()
        assert isinstance(tracer.reporter._sender, HTTPSender)
        tracer.close()



# Modified by SignalFx
# Copyright (c) 2016-2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import division

import math
import time

from six.moves import range
from six import PY3
import requests
import pytest
import mock

from jaeger_client.sampler import (
    Sampler,
    ConstSampler,
    ProbabilisticSampler,
    RateLimitingSampler,
    RemoteControlledSampler,
    GuaranteedThroughputProbabilisticSampler,
    AdaptiveSampler,
    DEFAULT_MAX_OPERATIONS,
    DEFAULT_SAMPLING_PROBABILITY,
    get_sampling_probability,
    get_rate_limit,
)

MAX_INT = 1 << 63


def get_tags(type, param):
    return {
        'sampler.type': type,
        'sampler.param': param,
    }


def test_abstract_sampler_errors():
    sampler = Sampler()
    with pytest.raises(NotImplementedError):
        sampler.is_sampled(trace_id=123)
    with pytest.raises(NotImplementedError):
        sampler.close()


def test_probabilistic_sampler_errors():
    with pytest.raises(AssertionError):
        ProbabilisticSampler(-0.1)
    with pytest.raises(AssertionError):
        ProbabilisticSampler(1.1)


def test_probabilistic_sampler():
    sampler = ProbabilisticSampler(0.5)
    assert MAX_INT == 0x8000000000000000
    sampled, tags = sampler.is_sampled(MAX_INT - 10)
    assert sampled
    assert tags == get_tags('probabilistic', 0.5)
    sampled, _ = sampler.is_sampled(MAX_INT + 10)
    assert not sampled
    sampler.close()
    assert '%s' % sampler == 'ProbabilisticSampler(0.5)'


def test_const_sampler():
    sampler = ConstSampler(True)
    sampled, _ = sampler.is_sampled(1)
    assert sampled
    sampled, _ = sampler.is_sampled(MAX_INT)
    assert sampled
    sampler = ConstSampler(False)
    sampled, tags = sampler.is_sampled(1)
    assert not sampled
    sampled, tags = sampler.is_sampled(MAX_INT)
    assert not sampled
    assert tags == get_tags('const', False)
    assert '%s' % sampler == 'ConstSampler(False)'


def test_rate_limiting_sampler():
    sampler = RateLimitingSampler(2)
    sampler.rate_limiter.balance = 2.0
    # stop time by overwriting timestamp() function to always return
    # the same time
    ts = time.time()
    sampler.rate_limiter.last_tick = ts
    with mock.patch('jaeger_client.rate_limiter.RateLimiter.timestamp') \
            as mock_time:
        mock_time.side_effect = lambda: ts  # always return same time
        assert sampler.rate_limiter.timestamp() == ts
        sampled, _ = sampler.is_sampled(0)
        assert sampled, 'initial balance allows first item'
        sampled, _ = sampler.is_sampled(0)
        assert sampled, 'initial balance allows second item'
        sampled, _ = sampler.is_sampled(0)
        assert not sampled, 'initial balance exhausted'

        # move time 250ms forward, not enough credits to pay for one sample
        mock_time.side_effect = lambda: ts + 0.25
        sampled, _ = sampler.is_sampled(0)
        assert not sampled, 'not enough time passed for full item'

        # move time 500ms forward, now enough credits to pay for one sample
        mock_time.side_effect = lambda: ts + 0.5
        sampled, _ = sampler.is_sampled(0)
        assert sampled, 'enough time for new item'
        sampled, _ = sampler.is_sampled(0)
        assert not sampled, 'no more balance'

        # move time 5s forward, enough to accumulate credits for 10 samples,
        # but it should still be capped at 2
        sampler.last_tick = ts  # reset the timer
        mock_time.side_effect = lambda: ts + 5
        sampled, _ = sampler.is_sampled(0)
        assert sampled, 'enough time for new item'
        sampled, _ = sampler.is_sampled(0)
        assert sampled, 'enough time for second new item'
        for i in range(0, 8):
            sampled, tags = sampler.is_sampled(0)
            assert not sampled, 'but no further, since time is stopped'
        assert tags == get_tags('ratelimiting', 2)
    sampler.close()
    assert '%s' % sampler == 'RateLimitingSampler(2)'

    # Test with rate limit of greater than 1 second
    sampler = RateLimitingSampler(0.1)
    sampler.rate_limiter.balance = 1.0
    ts = time.time()
    sampler.rate_limiter.last_tick = ts
    with mock.patch('jaeger_client.rate_limiter.RateLimiter.timestamp') \
            as mock_time:
        mock_time.side_effect = lambda: ts  # always return same time
        assert sampler.rate_limiter.timestamp() == ts
        sampled, _ = sampler.is_sampled(0)
        assert sampled, 'initial balance allows first item'
        sampled, _ = sampler.is_sampled(0)
        assert not sampled, 'initial balance exhausted'

        # move time 11s forward, enough credits to pay for one sample
        mock_time.side_effect = lambda: ts + 11
        sampled, _ = sampler.is_sampled(0)
        assert sampled
    sampler.close()
    assert '%s' % sampler == 'RateLimitingSampler(0.1)'

    # Test update
    sampler = RateLimitingSampler(3.0)
    sampler.rate_limiter.balance = 3.0
    ts = time.time()
    sampler.rate_limiter.last_tick = ts
    with mock.patch('jaeger_client.rate_limiter.RateLimiter.timestamp') \
            as mock_time:
        mock_time.side_effect = lambda: ts  # always return same time
        assert sampler.rate_limiter.timestamp() == ts
        sampled, _ = sampler.is_sampled(0)
        assert sampled
        assert sampler.rate_limiter.balance == 2.0
        assert '%s' % sampler == 'RateLimitingSampler(3.0)'

        sampler.update(3.0)
        assert '%s' % sampler == \
               'RateLimitingSampler(3.0)', 'should short cirtcuit if rate is the same'

        sampler.update(2.0)
        assert sampler.rate_limiter.balance == 4.0 / 3.0
        assert '%s' % sampler == 'RateLimitingSampler(2.0)'
    sampler.close()


def test_guaranteed_throughput_probabilistic_sampler():
    sampler = GuaranteedThroughputProbabilisticSampler('op',
                                                       2,
                                                       0.5)
    sampler.lower_bound_sampler.rate_limiter.balance = 2.0
    sampled, tags = sampler.is_sampled(MAX_INT - 10)
    assert sampled
    assert tags == get_tags('probabilistic', 0.5)
    sampled, tags = sampler.is_sampled(MAX_INT + 10)
    assert sampled
    assert tags == get_tags('lowerbound', 0.5)
    sampled, _ = sampler.is_sampled(MAX_INT + 10)
    assert not sampled
    assert '%s' % sampler == 'GuaranteedThroughputProbabilisticSampler(op, 0.500000, 2.000000)'

    sampler.update(3, 0.51)
    sampler.lower_bound_sampler.rate_limiter.balance = 3.0
    sampled, tags = sampler.is_sampled(MAX_INT - 10)
    assert sampled
    assert tags == get_tags('probabilistic', 0.51)
    sampled, tags = sampler.is_sampled(int(MAX_INT + (MAX_INT / 4)))
    assert sampled
    assert tags == get_tags('lowerbound', 0.51)

    assert '%s' % sampler == 'GuaranteedThroughputProbabilisticSampler(op, 0.510000, 3.000000)'
    sampler.close()


def test_adaptive_sampler():
    strategies = {
        'defaultSamplingProbability': 0.51,
        'defaultLowerBoundTracesPerSecond': 3,
        'perOperationStrategies':
        [
            {
                'operation': 'op',
                'probabilisticSampling': {
                    'samplingRate': 0.5
                }
            }
        ]
    }
    sampler = AdaptiveSampler(strategies, 2)
    sampled, tags = sampler.is_sampled(MAX_INT - 10, 'op')
    assert sampled
    assert tags == get_tags('probabilistic', 0.5)

    # This operation is seen for the first time by the sampler
    sampled, tags = sampler.is_sampled(MAX_INT - 10, 'new_op')
    assert sampled
    assert tags == get_tags('probabilistic', 0.51)

    ts = time.time()
    with mock.patch('jaeger_client.rate_limiter.RateLimiter.timestamp') \
            as mock_time:

        # Move time forward by a second to guarantee the rate limiter has enough credits
        mock_time.side_effect = lambda: ts + 1

        sampled, tags = sampler.is_sampled(int(MAX_INT + (MAX_INT / 4)), 'new_op')
        assert sampled
        assert tags == get_tags('lowerbound', 0.51)

    # This operation is seen for the first time by the sampler but surpasses
    # max_operations of 2. The default probabilistic sampler will be used
    sampled, tags = sampler.is_sampled(MAX_INT - 10, 'new_op_2')
    assert sampled
    assert tags == get_tags('probabilistic', 0.51)
    sampled, _ = sampler.is_sampled(int(MAX_INT + (MAX_INT / 4)), 'new_op_2')
    assert not sampled
    assert '%s' % sampler == 'AdaptiveSampler(0.510000, 3.000000, 2)'

    # Update the strategies
    strategies = {
        'defaultSamplingProbability': 0.52,
        'defaultLowerBoundTracesPerSecond': 4,
        'perOperationStrategies':
        [
            {
                'operation': 'op',
                'probabilisticSampling': {
                    'samplingRate': 0.52
                }
            },
            {
                'operation': 'new_op_3',
                'probabilisticSampling': {
                    'samplingRate': 0.53
                }
            }
        ]
    }
    sampler.update(strategies)

    # The probability for op has been updated
    sampled, tags = sampler.is_sampled(MAX_INT - 10, 'op')
    assert sampled
    assert tags == get_tags('probabilistic', 0.52)

    # A new operation has been added
    sampled, tags = sampler.is_sampled(MAX_INT - 10, 'new_op_3')
    assert sampled
    assert tags == get_tags('probabilistic', 0.53)
    assert '%s' % sampler == 'AdaptiveSampler(0.520000, 4.000000, 2)'

    sampler.close()


def test_adaptive_sampler_default_values():
    adaptive_sampler = AdaptiveSampler({}, 2)
    assert '%s' % adaptive_sampler == \
           'AdaptiveSampler(0.001000, 0.001667, 2)', 'sampler should use default values'

    sampled, tags = adaptive_sampler.is_sampled(0, 'op')
    assert sampled
    assert tags == \
        get_tags('probabilistic', 0.001), 'should use default probability'
    assert '%s' % adaptive_sampler.samplers['op'] == \
           'GuaranteedThroughputProbabilisticSampler(op, 0.001000, 0.001667)'

    adaptive_sampler.update(strategies={
        'defaultLowerBoundTracesPerSecond': 4,
        'perOperationStrategies':
            [
                {
                    'operation': 'new_op',
                    'probabilisticSampling': {
                        'samplingRate': 0.002
                    }
                }
            ]
    })
    assert '%s' % adaptive_sampler == 'AdaptiveSampler(0.001000, 4.000000, 2)'

    sampled, tags = adaptive_sampler.is_sampled(0, 'new_op')
    assert sampled
    assert tags == get_tags('probabilistic', 0.002)
    assert '%s' % adaptive_sampler.samplers['new_op'] == \
           'GuaranteedThroughputProbabilisticSampler(new_op, 0.002000, 4.000000)'

    sampled, tags = adaptive_sampler.is_sampled(0, 'op')
    assert sampled
    assert tags == get_tags('probabilistic', 0.001)
    # TODO ruh roh, the lowerbound isn't changed
    #  if the operation isn't included in perOperationStrategies
    assert '%s' % adaptive_sampler.samplers['op'] == \
           'GuaranteedThroughputProbabilisticSampler(op, 0.001000, 0.001667)'


def test_sampler_equality():
    const1 = ConstSampler(True)
    const2 = ConstSampler(True)
    const3 = ConstSampler(False)
    assert const1 == const2
    assert const1 != const3

    prob1 = ProbabilisticSampler(rate=0.01)
    prob2 = ProbabilisticSampler(rate=0.01)
    prob3 = ProbabilisticSampler(rate=0.02)
    assert prob1 == prob2
    assert prob1 != prob3
    assert const1 != prob1

    rate1 = RateLimitingSampler(max_traces_per_second=0.01)
    rate2 = RateLimitingSampler(max_traces_per_second=0.01)
    rate3 = RateLimitingSampler(max_traces_per_second=0.02)
    assert rate1 == rate2
    assert rate1 != rate3
    assert rate1 != const1
    assert rate1 != prob1


def test_remotely_controlled_sampler():
    sampler = RemoteControlledSampler(
        channel=mock.MagicMock(),
        service_name='x',
        sampling_refresh_interval=.001
    )
    sampled, tags = sampler.is_sampled(1)
    assert sampled
    assert tags == get_tags('probabilistic', DEFAULT_SAMPLING_PROBABILITY)
    sampler.close()

    init_sampler = mock.MagicMock()
    init_sampler.is_sampled = mock.MagicMock()
    channel = mock.MagicMock()
    sampler = RemoteControlledSampler(
        channel=channel,
        service_name='x',
        init_sampler=init_sampler,
        logger=mock.MagicMock(),
        sampling_refresh_interval=.001
    )
    assert init_sampler.is_sampled.call_count == 1

    sampler.is_sampled(1)
    assert init_sampler.is_sampled.call_count == 2

    sampler._create_periodic_callback = mock.MagicMock()
    # noinspection PyProtectedMember
    sampler._init_polling()

    # noinspection PyProtectedMember
    sampler._delayed_polling()
    sampler.close()

    sampler = RemoteControlledSampler(
        channel=mock.MagicMock(),
        service_name='x',
        max_operations=None,
        sampling_refresh_interval=.001
    )
    assert sampler.max_operations == DEFAULT_MAX_OPERATIONS

    sampler.close()
    assert not sampler.running
    sampler._init_polling()
    assert not sampler.running
    sampler._delayed_polling()
    assert not sampler.running


class MockResponse(requests.Response):
    """
    Helps determine whether sampling strategies are successfully parsed by requests
    and any errors are properly handled.
    """
    def __init__(self, content):
        super(MockResponse, self).__init__()
        self._content = content
        if PY3:
            self.encoding = 'UTF-8'
            if not isinstance(content, (bytearray, bytes, type(None))):
                self._content = bytes(content, 'utf-8')


# noinspection PyProtectedMember
def test_sampling_request_callback():
    channel = mock.MagicMock()
    error_reporter = mock.MagicMock()
    error_reporter.error = mock.MagicMock()
    sampler = RemoteControlledSampler(
        channel=channel,
        service_name='x',
        error_reporter=error_reporter,
        max_operations=10,
    )

    probabilistic_strategy = """
    {
        "strategyType":"PROBABILISTIC",
        "probabilisticSampling":
        {
            "samplingRate":0.002
        }
    }
    """
    response = MockResponse(probabilistic_strategy)
    sampler._sampling_request_callback(response, None)
    assert '%s' % sampler.sampler == \
           'ProbabilisticSampler(0.002)', 'sampler should have changed to probabilistic'
    prev_sampler = sampler.sampler

    sampler._sampling_request_callback(response, None)
    assert prev_sampler is sampler.sampler, \
        "strategy hasn't changed so sampler should not change"

    adaptive_sampling_strategy = """
    {
        "strategyType":"PROBABILISTIC",
        "operationSampling":
        {
            "defaultSamplingProbability":0.001,
            "defaultLowerBoundTracesPerSecond":2,
            "perOperationStrategies":
            [
                {
                    "operation":"op",
                    "probabilisticSampling":{
                        "samplingRate":0.002
                    }
                }
            ]
        }
    }
    """
    response = MockResponse(adaptive_sampling_strategy)
    sampler._sampling_request_callback(response, None)
    assert '%s' % sampler.sampler == 'AdaptiveSampler(0.001000, 2.000000, 10)', \
        'sampler should have changed to adaptive'
    prev_sampler = sampler.sampler

    sampler._sampling_request_callback(response, None)
    assert prev_sampler is sampler.sampler, "strategy hasn't changed so sampler should not change"

    response = MockResponse(probabilistic_strategy.encode('utf-8'))
    sampler._sampling_request_callback(response, None)
    assert '%s' % sampler.sampler == \
           'ProbabilisticSampler(0.002)', 'sampler should have changed to probabilistic'

    response = MockResponse(bytearray(adaptive_sampling_strategy.encode('utf-8')))
    sampler._sampling_request_callback(response, None)
    assert '%s' % sampler.sampler == 'AdaptiveSampler(0.001000, 2.000000, 10)', \
        'sampler should have changed to adaptive'
    prev_sampler = sampler.sampler

    sampler._sampling_request_callback(response, Exception())
    assert error_reporter.error.call_count == 1
    assert prev_sampler is sampler.sampler, 'error fetching strategy should not update the sampler'

    response = MockResponse('bad_json')
    sampler._sampling_request_callback(response, None)
    assert error_reporter.error.call_count == 2
    assert prev_sampler is sampler.sampler, 'error updating sampler should not update the sampler'

    response = MockResponse(None)
    sampler._sampling_request_callback(response, None)
    assert error_reporter.error.call_count == 3
    assert prev_sampler is sampler.sampler, 'error updating sampler should not update the sampler'

    response = MockResponse(probabilistic_strategy)
    sampler._sampling_request_callback(response, None)
    assert '%s' % sampler.sampler == 'ProbabilisticSampler(0.002)', \
        'updating sampler from adaptive to probabilistic should work'
    assert error_reporter.error.call_count == 3

    sampler.close()


probabilistic_sampler = ProbabilisticSampler(0.002)
other_probabilistic_sampler = ProbabilisticSampler(0.003)
rate_limiting_sampler = RateLimitingSampler(10)
other_rate_limiting_sampler = RateLimitingSampler(20)


@pytest.mark.parametrize(
    'response,init_sampler,expected_sampler,err_count,err_msg,reference_equivalence,max_operations',
    [
        (
            {'strategyType': 'PROBABILISTIC', 'probabilisticSampling': {'samplingRate': 0.003}},
            probabilistic_sampler,
            other_probabilistic_sampler,
            0,
            'sampler should update to new probabilistic sampler',
            False,
            10,
        ),
        (
            {'strategyType': 'PROBABILISTIC', 'probabilisticSampling': {'samplingRate': 400}},
            probabilistic_sampler,
            probabilistic_sampler,
            1,
            'sampler should remain the same if strategy is invalid',
            True,
            10,
        ),
        (
            {'strategyType': 'PROBABILISTIC', 'probabilisticSampling': {'samplingRate': 0.002}},
            probabilistic_sampler,
            probabilistic_sampler,
            0,
            'sampler should remain the same with the same strategy',
            True,
            10,
        ),
        (
            {'strategyType': 'RATE_LIMITING', 'rateLimitingSampling': {'maxTracesPerSecond': 10}},
            probabilistic_sampler,
            rate_limiting_sampler,
            0,
            'sampler should update to new rate limiting sampler',
            False,
            10,
        ),
        (
            {'strategyType': 'RATE_LIMITING', 'rateLimitingSampling': {'maxTracesPerSecond': 10}},
            rate_limiting_sampler,
            rate_limiting_sampler,
            0,
            'sampler should remain the same with the same strategy',
            True,
            10,
        ),
        (
            {'strategyType': 'RATE_LIMITING', 'rateLimitingSampling': {'maxTracesPerSecond': -10}},
            rate_limiting_sampler,
            rate_limiting_sampler,
            1,
            'sampler should remain the same if strategy is invalid',
            True,
            10,
        ),
        (
            {'strategyType': 'RATE_LIMITING', 'rateLimitingSampling': {'maxTracesPerSecond': 20}},
            rate_limiting_sampler,
            other_rate_limiting_sampler,
            0,
            'sampler should update to new rate limiting sampler',
            False,
            10,
        ),
        (
            {},
            rate_limiting_sampler,
            rate_limiting_sampler,
            1,
            'sampler should remain the same if strategy is empty',
            True,
            10,
        ),
        (
            {'strategyType': 'INVALID_TYPE'},
            rate_limiting_sampler,
            rate_limiting_sampler,
            1,
            'sampler should remain the same if strategy is invalid',
            True,
            10,
        ),
        (
            {'strategyType': 'INVALID_TYPE'},
            rate_limiting_sampler,
            rate_limiting_sampler,
            1,
            'sampler should remain the same if strategy is invalid',
            True,
            None,
        ),
    ]
)
def test_update_sampler(response, init_sampler, expected_sampler,
                        err_count, err_msg, reference_equivalence, max_operations):
    error_reporter = mock.MagicMock()
    error_reporter.error = mock.MagicMock()
    remote_sampler = RemoteControlledSampler(
        channel=mock.MagicMock(),
        service_name='x',
        error_reporter=error_reporter,
        max_operations=max_operations,
        init_sampler=init_sampler,
    )

    # noinspection PyProtectedMember
    remote_sampler._update_sampler(response)
    assert error_reporter.error.call_count == err_count
    if reference_equivalence:
        assert remote_sampler.sampler is expected_sampler, err_msg
    else:
        assert remote_sampler.sampler == expected_sampler, err_msg

    remote_sampler.close()


# noinspection PyProtectedMember
def test_update_sampler_adaptive_sampler():
    error_reporter = mock.MagicMock()
    error_reporter.error = mock.MagicMock()
    remote_sampler = RemoteControlledSampler(
        channel=mock.MagicMock(),
        service_name='x',
        error_reporter=error_reporter,
        max_operations=10,
    )

    response = {
        'strategyType': 'RATE_LIMITING',
        'operationSampling':
        {
            'defaultSamplingProbability': 0.001,
            'defaultLowerBoundTracesPerSecond': 2,
            'perOperationStrategies':
            [
                {
                    'operation': 'op',
                    'probabilisticSampling': {
                        'samplingRate': 0.002
                    }
                }
            ]
        }
    }

    remote_sampler._update_sampler(response)
    assert '%s' % remote_sampler.sampler == 'AdaptiveSampler(0.001000, 2.000000, 10)'

    new_response = {
        'strategyType': 'RATE_LIMITING',
        'operationSampling':
        {
            'defaultSamplingProbability': 0.51,
            'defaultLowerBoundTracesPerSecond': 3,
            'perOperationStrategies':
            [
                {
                    'operation': 'op',
                    'probabilisticSampling': {
                        'samplingRate': 0.002
                    }
                }
            ]
        }
    }

    remote_sampler._update_sampler(new_response)
    assert '%s' % remote_sampler.sampler == 'AdaptiveSampler(0.510000, 3.000000, 10)'

    remote_sampler._update_sampler(
        {'strategyType': 'PROBABILISTIC', 'probabilisticSampling': {'samplingRate': 0.004}})
    assert '%s' % remote_sampler.sampler == 'ProbabilisticSampler(0.004)', \
        'should not fail going from adaptive sampler to probabilistic sampler'

    remote_sampler._update_sampler({'strategyType': 'RATE_LIMITING',
                                    'operationSampling': {'defaultSamplingProbability': 0.4}})
    assert '%s' % remote_sampler.sampler == 'AdaptiveSampler(0.400000, 0.001667, 10)'

    remote_sampler.close()


@pytest.mark.parametrize('strategy,expected', [
    ({'probabilisticSampling': {'samplingRate': 0.003}}, 0.003),
    ({}, 0.001),
    (None, 0.001),
    ({'probabilisticSampling': {}}, 0.001),
    ({'probabilisticSampling': None}, 0.001),
])
def test_get_sampling_probability(strategy, expected):
    assert expected == get_sampling_probability(strategy)


@pytest.mark.parametrize('strategy,expected', [
    ({'rateLimitingSampling': {'maxTracesPerSecond': 1}}, 1),
    ({}, 0.0016666),
    (None, 0.0016666),
    ({'rateLimitingSampling': {}}, 0.0016666),
    ({'rateLimitingSampling': None}, 0.0016666),
])
def test_get_rate_limit(strategy, expected):
    assert math.fabs(expected - get_rate_limit(strategy)) < 0.0001



# Copyright (c) 2017 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from six.moves import range
import time
import mock

from jaeger_client.rate_limiter import RateLimiter


def test_rate_limiting_sampler():
    rate_limiter = RateLimiter(2, 2)
    assert rate_limiter.balance <= 2.0
    rate_limiter.balance = 2.0
    # stop time by overwriting timestamp() function to always return
    # the same time
    ts = time.time()
    rate_limiter.last_tick = ts
    with mock.patch('jaeger_client.rate_limiter.RateLimiter.timestamp') \
            as mock_time:
        mock_time.side_effect = lambda: ts  # always return same time
        assert rate_limiter.timestamp() == ts
        assert rate_limiter.check_credit(1), 'initial balance allows first item'
        assert rate_limiter.check_credit(1), 'initial balance allows second item'
        assert not rate_limiter.check_credit(1), 'initial balance exhausted'

        # move time 250ms forward, not enough credits to pay for one sample
        mock_time.side_effect = lambda: ts + 0.25
        assert not rate_limiter.check_credit(1), 'not enough time passed for full item'

        # move time 500ms forward, now enough credits to pay for one sample
        mock_time.side_effect = lambda: ts + 0.5
        assert rate_limiter.check_credit(1), 'enough time for new item'
        assert not rate_limiter.check_credit(1), 'no more balance'

        # move time 5s forward, enough to accumulate credits for 10 samples,
        # but it should still be capped at 2
        rate_limiter.last_tick = ts  # reset the timer
        mock_time.side_effect = lambda: ts + 5
        assert rate_limiter.check_credit(1), 'enough time for new item'
        assert rate_limiter.check_credit(1), 'enough time for second new item'
        for i in range(0, 8):
            assert not rate_limiter.check_credit(1), 'but no further, since time is stopped'

    # Test with rate limit of greater than 1 second
    rate_limiter = RateLimiter(0.1, 1.0)
    assert rate_limiter.balance <= 1.0
    rate_limiter.balance = 1.0
    ts = time.time()
    rate_limiter.last_tick = ts
    with mock.patch('jaeger_client.rate_limiter.RateLimiter.timestamp') \
            as mock_time:
        mock_time.side_effect = lambda: ts  # always return same time
        assert rate_limiter.timestamp() == ts
        assert rate_limiter.check_credit(1), 'initial balance allows first item'
        assert not rate_limiter.check_credit(1), 'initial balance exhausted'

        # move time 11s forward, enough credits to pay for one sample
        mock_time.side_effect = lambda: ts + 11
        assert rate_limiter.check_credit(1)

    # Test update
    rate_limiter = RateLimiter(3.0, 3.0)
    assert rate_limiter.balance <= 3.0
    rate_limiter.balance = 3.0
    ts = time.time()
    rate_limiter.last_tick = ts
    with mock.patch('jaeger_client.rate_limiter.RateLimiter.timestamp') \
            as mock_time:
        mock_time.side_effect = lambda: ts  # always return same time
        assert rate_limiter.timestamp() == ts
        assert rate_limiter.check_credit(1)
        rate_limiter.update(2.0, 2.0)
        assert rate_limiter.balance == 4.0 / 3.0



# Modified by SignalFx
# Copyright (c) 2016 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import mock
import unittest

from jaeger_client import utils


class ConfigTests(unittest.TestCase):

    def check_boolean(self, string, default, correct):
        assert utils.get_boolean(string, default) == correct

    def test_get_false_boolean(self):
        self.check_boolean('false', 'asdf', False)

    def test_get_0_boolean(self):
        self.check_boolean('0', 'asdf', False)

    def test_get_true_boolean(self):
        self.check_boolean('true', 'qwer', True)

    def test_get_1_boolean(self):
        self.check_boolean('1', 'qwer', True)

    def test_get_unknown_boolean(self):
        self.check_boolean('zxcv', 'qwer', 'qwer')

    def test_get_None_boolean(self):
        self.check_boolean(None, 'qwer', False)

#    def test_error_reporter_doesnt_send_metrics_if_not_configured(self):
#        er = utils.ErrorReporter(False)
#        er.error('foo', 1)
#        assert not mock_metrics.count.called

    def test_error_reporter_sends_metrics_if_configured(self):
        mock_metrics = mock.MagicMock()
        er = utils.ErrorReporter(mock_metrics)
        er.error('foo', 1)
        assert mock_metrics.count.called_with('foo', 1)

    def test_error_reporter_doesnt_send_log_messages_if_before_deadline(self):
        mock_logger = mock.MagicMock()
        er = utils.ErrorReporter(None, logger=mock_logger, log_interval_minutes=1000)
        er.error('foo', 1)
        assert not mock_logger.error.called

    def test_error_reporter_sends_log_messages_if_after_deadline(self):
        mock_logger = mock.MagicMock()
        # 0 log interval means we're always after the deadline, so always log
        er = utils.ErrorReporter(None, logger=mock_logger, log_interval_minutes=0)
        er._last_error_reported_at = 0
        er.error('foo', 1, 'error args')
        assert mock_logger.error.call_args == (('foo', 1, 'error args',),)


def test_local_ip_does_not_blow_up():
    import socket
    import jaeger_client.utils
    socket.gethostname()
    with mock.patch('socket.gethostbyname',
                    side_effect=[IOError(), '127.0.0.1']):
        jaeger_client.utils.local_ip()


def test_get_local_ip_by_socket_does_not_blow_up():
    import jaeger_client.utils
    jaeger_client.utils.get_local_ip_by_socket()


class CustomException(Exception):
    pass


def test_raise_with_value_reraises():
    try:
        raise CustomException('Initial Value')
    except Exception as e:
        try:
            utils.raise_with_value(e, 'New Value')
        except Exception as e:
            assert type(e) == CustomException
            assert str(e) == 'New Value'
        else:
            assert False, 'Failed to raise'
    else:
        assert False, 'Failed to raise'


def test_raise_with_value_raises():
    try:
        utils.raise_with_value(CustomException, 'Some Value')
    except Exception as e:
        assert type(e) == CustomException
        assert str(e) == 'Some Value'
    else:
        assert False, 'Failed to raise'



# Copyright (c) 2018 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import

import socket
import logging

import requests
import requests.auth
from . import thrift
from .utils import raise_with_value
from .local_agent_net import LocalAgentSender
from thrift.TSerialization import serialize
from thrift.protocol import TCompactProtocol
from thrift.transport import TTransport

from jaeger_client.thrift_gen.agent import Agent


logger = logging.getLogger('jaeger_tracing')


class Sender(object):
    def __init__(self, batch_size=10):
        """
        :param batch_size: how many spans we can submit at once to Collector
        """
        from threading import Lock
        self._process_lock = Lock()
        self._process = None
        self.batch_size = batch_size
        self.spans = []

    def append(self, span):
        """
        Queue a span for subsequent submission calls to flush().
        If number of appended spans is equal to batch size, initiate flush().
        """
        spans_flushed = 0
        self.spans.append(span)

        if len(self.spans) == self.batch_size:
            spans_flushed = self.flush()

        return spans_flushed

    @property
    def span_count(self):
        return len(self.spans)

    def flush(self):
        """
        Flush spans, if any, if process has been set. Returns number of spans successfully flushed.
        """
        spans_sent = 0
        if self.spans:
            with self._process_lock:
                process = self._process
            if process:
                try:
                    spans_sent = self._batch_and_send(self.spans, self._process)
                finally:
                    self.spans = []
        return spans_sent

    def _batch_and_send(self, spans, process):
        """
        Batch spans and invokes send(), returning number of spans sent.
        Override with specific batching logic, if desired.
        """
        batch = thrift.make_jaeger_batch(spans=spans, process=process)
        self.send(batch)
        return len(spans)

    def send(self, batch):
        """Send batch of spans to collector via desired transport."""
        raise NotImplementedError('This method should be implemented by subclasses')

    def set_process(self, service_name, tags, max_length):
        with self._process_lock:
            self._process = thrift.make_process(
                service_name=service_name, tags=tags, max_length=max_length,
            )

    def getProtocol(self, transport):
        raise NotImplementedError('This method should be implemented by subclasses')


class UDPSenderException(Exception):

    pass


class UDPSender(Sender):

    def __init__(self, channel=None, batch_size=10):
        super(UDPSender, self).__init__(batch_size=batch_size)
        self._channel = channel or self._create_local_agent_channel()
        self._host = self._channel._host
        self._port = self._channel._reporting_port
        self._agent = Agent.Client(self.getProtocol(self._channel))
        self._max_span_space = None

    def send(self, batch):
        """
        Send batch of spans out via thrift transport.
        """
        try:
            self._agent.emitBatch(batch)
        except socket.error as e:
            raise_with_value(e, 'Failed to submit traces to jaeger-agent socket: {}'.format(e))
        except Exception as e:
            raise_with_value(e, 'Failed to submit traces to jaeger-agent: {}'.format(e))

    def _batch_and_send(self, spans, process):
        """
        Batches and sends spans in as many UDP packets as necessary.  Will drop any span with size
        greater than allowable relative to minimum batch size and maximum UDP payload.
        """
        flush_error = None
        batched_spans = []
        total_span_bytes = 0
        if self._max_span_space is None:
            self._max_span_space = 65000 - self._calculate_base_batch_size(process)

        for span in spans:
            span_size = self._calculate_span_size(span)
            if span_size > self._max_span_space:
                flush_error = UDPSenderException(
                    'Cannot send span of size {}. Dropping.'.format(span_size)
                )
            elif total_span_bytes + span_size > self._max_span_space:
                batch = thrift.make_jaeger_batch(spans=batched_spans, process=process)
                self.send(batch)
                batched_spans = [span]
                total_span_bytes = span_size
            else:
                batched_spans.append(span)
                total_span_bytes += span_size

        batch = thrift.make_jaeger_batch(spans=batched_spans, process=process)
        self.send(batch)

        if flush_error is not None:
            raise flush_error

        return len(batched_spans)

    def _calculate_base_batch_size(self, process):
        """Determine what size the batch will be without any spans"""
        buff = TTransport.TMemoryBuffer()
        proto = self.getProtocol(buff)
        base_batch = thrift.make_jaeger_batch(spans=[], process=process)
        base_batch.write(proto)
        return len(buff.getvalue())

    def _calculate_span_size(self, span):
        buff = TTransport.TMemoryBuffer()
        proto = self.getProtocol(buff)
        jaeger_span = thrift.make_jaeger_span(span)
        jaeger_span.write(proto)
        return len(buff.getvalue())

    def _create_local_agent_channel(self):
        """
        Create an out-of-process channel communicating to local jaeger-agent.
        Spans are submitted as SOCK_DGRAM Thrift, sampling strategy is polled
        via JSON HTTP.

        :param self: instance of Config
        """
        from .config import DEFAULT_REPORTING_HOST, DEFAULT_REPORTING_PORT, DEFAULT_SAMPLING_PORT
        return LocalAgentSender(
            host=DEFAULT_REPORTING_HOST,
            sampling_port=DEFAULT_SAMPLING_PORT,
            reporting_port=DEFAULT_REPORTING_PORT
        )

    # method for protocol factory
    def getProtocol(self, transport):
        """
        Implements Thrift ProtocolFactory interface
        :param: transport:
        :return: Thrift compact protocol
        """
        return TCompactProtocol.TCompactProtocol(transport)


class HTTPSender(Sender):
    def __init__(self, endpoint, auth_token='', user='', password='', batch_size=10):
        super(HTTPSender, self).__init__(batch_size=batch_size)
        self.client = requests.Session()
        self.url = endpoint
        self.auth_token = auth_token
        self.user = user
        self.password = password
        if any((self.user, self.password)):
            self.client.auth = requests.auth.HTTPBasicAuth(self.user, self.password)

    def send(self, batch):
        """
        Send batch of spans out via AsyncHTTPClient. Any exceptions thrown
        will be caught above in the exception handler of _submit().
        """
        headers = {'Content-Type': 'application/x-thrift'}
        if self.auth_token:
            headers['Authorization'] = 'Bearer {}'.format(self.auth_token)

        data = serialize(batch)
        headers['Content-Length'] = str(len(data))

        try:
            self.client.post(url=self.url, headers=headers, data=data)
        except Exception as e:
            raise_with_value(e, 'POST to jaeger_endpoint failed: {}'.format(e))



# Modified by SignalFx
# Copyright (c) 2016 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

import time
import logging
import threading

from .constants import DEFAULT_FLUSH_INTERVAL
from .metrics import Metrics, LegacyMetricsFactory
from .utils import ErrorReporter
from six.moves import queue


default_logger = logging.getLogger('jaeger_tracing')


class NullReporter(object):
    """Ignores all spans."""
    def report_span(self, span):
        pass

    def set_process(self, service_name, tags, max_length):
        pass

    def close(self):
        return True


class InMemoryReporter(NullReporter):
    """Stores spans in memory and returns them via get_spans()."""
    def __init__(self):
        super(InMemoryReporter, self).__init__()
        self.spans = []
        self.lock = threading.Lock()

    def report_span(self, span):
        with self.lock:
            self.spans.append(span)

    def get_spans(self):
        with self.lock:
            return self.spans[:]


class LoggingReporter(NullReporter):
    """Logs all spans."""
    def __init__(self, logger=None):
        self.logger = logger if logger else default_logger

    def report_span(self, span):
        self.logger.info('Reporting span %s', span)


class Reporter(NullReporter):
    """Receives completed spans from Tracer and submits them out of process."""
    def __init__(self, sender=None, queue_capacity=100,
                 flush_interval=DEFAULT_FLUSH_INTERVAL, error_reporter=None,
                 metrics=None, metrics_factory=None, **kwargs):
        """
        :param sender: an instance of a senders.Sender subclass for sending
            batches of spans to jaeger.
        :param queue_capacity: how many spans we can hold in memory before
            starting to drop spans
        :param flush_interval: how often the auto-flush is called (in seconds)
        :param error_reporter:
        :param metrics: an instance of Metrics class, or None. This parameter
            has been deprecated, please use metrics_factory instead.
        :param metrics_factory: an instance of MetricsFactory class, or None.
        :param kwargs:
            'logger'
        :return:
        """
        from threading import Lock

        # TODO for next major rev: remove channel param in favor of sender
        # self.agent = Agent.Client()
        self._sender = sender
        self.queue_capacity = queue_capacity
        self.metrics_factory = metrics_factory or LegacyMetricsFactory(metrics or Metrics())
        self.metrics = ReporterMetrics(self.metrics_factory)
        self.error_reporter = error_reporter or ErrorReporter(Metrics())
        self.logger = kwargs.get('logger', default_logger)

        if queue_capacity < self._sender.batch_size:
            raise ValueError('Queue capacity cannot be less than sender batch size')

        self.queue = queue.Queue(maxsize=queue_capacity)
        self.stop = object()  # sentinel
        self.stopped = False
        self.stop_lock = Lock()
        self.flush_interval = flush_interval or None
        self._min_flush_interval = 10 ** -9
        self._consume_queue_thread = None

    def set_process(self, service_name, tags, max_length):
        self._sender.set_process(service_name, tags, max_length)

    def report_span(self, span):
        """
        Add a finished span to the queue, creating a `self._consume_queue()` worker thread
        if none exists.  This deferred creation of the consumer thread allows for easy adoption
        in forking environments (e.g. Django) where background threads shouldn't be spawned
        until the application is actively traced.
        """
        if self._consume_queue_thread is None:
            self._consume_queue_thread = threading.Thread(target=self._consume_queue)
            self._consume_queue_thread.daemon = True
            self._consume_queue_thread.start()

        try:
            with self.stop_lock:
                stopped = self.stopped
            if stopped:
                self.metrics.reporter_dropped(1)
            else:
                self.queue.put_nowait(span)
        except queue.Full:
            self.metrics.reporter_dropped(1)

    def _consume_queue(self):
        stopped = False
        while not stopped:
            interval = self.flush_interval
            spans_appended = 0
            while True:
                t0 = time.time()
                timeout = max(interval, self._min_flush_interval) if self.flush_interval else None
                try:
                    span = self.queue.get(timeout=timeout)
                    t1 = time.time()
                except queue.Empty:
                    # Always flush on timeouts, as they signify interval completion
                    break

                if span == self.stop:
                    stopped = True
                    self.queue.task_done()
                    break

                spans_appended += 1
                try:
                    spans_reported = self._sender.append(span)
                except Exception as exc:
                    # Assume append() triggered flush(), which failed and
                    # included all previously appended spans
                    self.metrics.reporter_failure(spans_appended)
                    self.error_reporter.error(exc)
                    for _ in range(spans_appended):
                        self.queue.task_done()
                    spans_appended = 0
                else:
                    if spans_reported:
                        self.metrics.reporter_success(spans_reported)
                        for _ in range(spans_reported):
                            self.queue.task_done()
                        spans_appended -= spans_reported

                if self.flush_interval:
                    interval -= t1 - t0

            try:
                try:
                    spans_reported = self._sender.flush()
                except Exception as exc:
                    # Assume number of failed spans is equal to previously appended
                    self.metrics.reporter_failure(spans_appended)
                    self.error_reporter.error(exc)
                    for _ in range(spans_appended):
                        self.queue.task_done()
                else:
                    self.metrics.reporter_success(spans_reported)
                    for _ in range(spans_reported):
                        self.queue.task_done()
                    self.metrics.reporter_queue_length(self.queue.qsize())
            except Exception as exc:
                self.logger.error(exc)
                break

        self.logger.info('Span publisher exited')

    def close(self):
        """
        Ensure that all spans from the queue are submitted.
        """
        with self.stop_lock:
            self.stopped = True
        if self._consume_queue_thread is not None:
            if self._consume_queue_thread.is_alive():
                self.queue.put(self.stop)
                self.queue.join()
            self._consume_queue_thread.join()

        return True


class ReporterMetrics(object):
    """Reporter specific metrics."""

    def __init__(self, metrics_factory):
        self.reporter_success = \
            metrics_factory.create_counter(name='jaeger:reporter_spans', tags={'result': 'ok'})
        self.reporter_failure = \
            metrics_factory.create_counter(name='jaeger:reporter_spans', tags={'result': 'err'})
        self.reporter_dropped = \
            metrics_factory.create_counter(name='jaeger:reporter_spans', tags={'result': 'dropped'})
        self.reporter_queue_length = \
            metrics_factory.create_gauge(name='jaeger:reporter_queue_length')


class CompositeReporter(NullReporter):
    """Delegates reporting to one or more underlying reporters."""
    def __init__(self, *reporters):
        self.reporters = reporters

    def set_process(self, service_name, tags, max_length):
        for reporter in self.reporters:
            reporter.set_process(service_name, tags, max_length)

    def report_span(self, span):
        for reporter in self.reporters:
            reporter.report_span(span)

    def close(self):
        closed = [reporter.close() for reporter in self.reporters]
        return all(closed)



# Modified by SignalFx
# Copyright (c) 2016 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import socket
import struct
import time
import sys

from six import reraise
from six.moves import range


class ErrorReporter(object):
    """
    Reports errors by emitting metrics, and if logger is provided,
    logging the error message once every log_interval_minutes

    N.B. metrics will be deprecated in the future
    """

    def __init__(self, metrics, logger=None, log_interval_minutes=15):
        self.logger = logger
        self.log_interval_minutes = log_interval_minutes
        self._last_error_reported_at = time.time()

    def error(self, *args):
        if self.logger is None:
            return

        next_logging_deadline = \
            self._last_error_reported_at + (self.log_interval_minutes * 60)
        current_time = time.time()
        if next_logging_deadline >= current_time:
            # If we aren't yet at the next logging deadline
            return

        self.logger.error(*args)
        self._last_error_reported_at = current_time


def get_boolean(string, default):
    string = str(string).lower()
    if string in ['false', '0', 'none']:
        return False
    elif string in ['true', '1']:
        return True
    else:
        return default


def local_ip():
    """Get the local network IP of this machine"""
    try:
        ip = socket.gethostbyname(socket.gethostname())
    except IOError:
        ip = socket.gethostbyname('localhost')
    if ip.startswith('127.'):
        ip = get_local_ip_by_interfaces()
        if ip is None:
            ip = get_local_ip_by_socket()
    return ip


def get_local_ip_by_socket():
    # Explanation : https://stackoverflow.com/questions/166506
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        # doesn't even have to be reachable
        s.connect(('10.255.255.255', 1))
        ip = s.getsockname()[0]
    except IOError:
        ip = None
    finally:
        s.close()
    return ip


def get_local_ip_by_interfaces():
    ip = None
    # Check eth0, eth1, eth2, en0, ...
    interfaces = [
        i + bytes(n) for i in (b'eth', b'en', b'wlan') for n in range(3)
    ]  # :(
    for interface in interfaces:
        try:
            ip = interface_ip(interface)
            if ip is not None:
                break
        except IOError:
            pass
    return ip


def interface_ip(interface):
    try:
        import fcntl
        """Determine the IP assigned to us by the given network interface."""
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        return socket.inet_ntoa(
            fcntl.ioctl(
                sock.fileno(), 0x8915, struct.pack('256s', interface[:15])
            )[20:24]
        )
    except ImportError:
        return None
    # Explanation:
    # http://stackoverflow.com/questions/11735821/python-get-localhost-ip
    # http://stackoverflow.com/questions/24196932/how-can-i-get-the-ip-address-of-eth0-in-python


def raise_with_value(exception, value):
    """
    Raise an instance of an exception type or reraise an instance with a
    modified message and unchanged traceback.
    """
    _type = exception if isinstance(exception, type) else type(exception)
    reraise(_type, _type(value), sys.exc_info()[2])



# Modified by SignalFx
# Copyright (c) 2016 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function, absolute_import
from six.moves import range

import logging
from threading import Event
import time
import collections

import mock
import jaeger_client.reporter

from jaeger_client import Span, SpanContext
from jaeger_client.metrics import LegacyMetricsFactory, Metrics
from jaeger_client.utils import ErrorReporter
from jaeger_client.reporter import Reporter
from jaeger_client.senders import Sender
from .conftest import wait_for


def test_null_reporter():
    reporter = jaeger_client.reporter.NullReporter()
    reporter.report_span({})
    assert reporter.close() is True


def test_in_memory_reporter():
    reporter = jaeger_client.reporter.InMemoryReporter()
    reporter.report_span({})
    assert reporter.close() is True
    spans = reporter.get_spans()
    assert [{}] == spans


def test_logging_reporter():
    log_mock = mock.MagicMock()
    reporter = jaeger_client.reporter.LoggingReporter(logger=log_mock)
    reporter.report_span({})
    log_mock.info.assert_called_with('Reporting span %s', {})
    assert reporter.close() is True


def test_composite_reporter():
    reporter = jaeger_client.reporter.CompositeReporter(
        jaeger_client.reporter.NullReporter(),
        jaeger_client.reporter.LoggingReporter())
    with mock.patch('jaeger_client.reporter.NullReporter.set_process') \
            as null_mock:
        with mock.patch('jaeger_client.reporter.LoggingReporter.set_process') \
                as log_mock:
            reporter.set_process('x', {}, 123)
            null_mock.assert_called_with('x', {}, 123)
            log_mock.assert_called_with('x', {}, 123)
    with mock.patch('jaeger_client.reporter.NullReporter.report_span') \
            as null_mock:
        with mock.patch('jaeger_client.reporter.LoggingReporter.report_span') \
                as log_mock:
            reporter.report_span({})
            null_mock.assert_called_with({})
            log_mock.assert_called_with({})
    with mock.patch('jaeger_client.reporter.NullReporter.close') \
            as null_mock:
        with mock.patch('jaeger_client.reporter.LoggingReporter.close') \
                as log_mock:

            null_mock.return_value = True
            log_mock.return_value = True
            assert reporter.close()
            null_mock.assert_called_once()
            log_mock.assert_called_once()


class FakeSender(object):
    """
    Mock the send() method of the reporter's Sender by capturing requests
    and returning incomplete futures that can be completed from
    inside the test.
    """
    def __init__(self):
        self.requests = []
        self.futures = []

    def __call__(self, batch):
        self.requests.append(batch)
        fut = Event()
        self.futures.append(fut)

        def waiter():
            fut.wait()
            return 1

        return waiter()


class HardErrorReporter(object):
    def error(self, *args):
        raise ValueError(*args)


FakeTrace = collections.namedtuple(
    'FakeTracer', ['ip_address', 'service_name'])


class FakeMetricsFactory(LegacyMetricsFactory):
    def __init__(self):
        super(FakeMetricsFactory, self).__init__(
            Metrics(count=self._incr_count)
        )
        self.counters = {}

    def _incr_count(self, key, value):
        self.counters[key] = value + self.counters.get(key, 0)


class TestReporter(object):

    @staticmethod
    def _new_span(name):
        tracer = FakeTrace(ip_address='127.0.0.1',
                           service_name='reporter_test')
        ctx = SpanContext(trace_id=1,
                          span_id=1,
                          parent_id=None,
                          flags=1)
        span = Span(context=ctx,
                    tracer=tracer,
                    operation_name=name)
        span.start_time = time.time()
        span.end_time = span.start_time + 0.001  # 1ms
        return span

    @staticmethod
    def _new_reporter(request, batch_size, flush=None, queue_cap=100, send_method=None):
        reporter = Reporter(sender=Sender(batch_size),
                            flush_interval=flush,
                            metrics_factory=FakeMetricsFactory(),
                            error_reporter=HardErrorReporter(),
                            queue_capacity=queue_cap)
        reporter.set_process('service', {}, max_length=0)
        send_method = send_method or FakeSender()
        reporter._sender.send = send_method
        request.addfinalizer(reporter.close)
        return reporter, send_method

    def test_submit_batch_size_1(self, request):
        reporter, sender = self._new_reporter(request, batch_size=1)

        reporter.report_span(self._new_span('1'))

        wait_for(lambda: len(sender.futures) > 0)
        assert 1 == len(sender.futures)

        sender.futures[0].set()
        reporter.close()
        assert 1 == len(sender.futures)

        # send after close
        span_dropped_key = 'jaeger:reporter_spans.result_dropped'
        assert span_dropped_key not in reporter.metrics_factory.counters
        reporter.report_span(self._new_span('1'))
        assert 1 == reporter.metrics_factory.counters[span_dropped_key]
        assert 1 == len(sender.futures)

    def test_sender_flushed_with_partial_batch(self, request):
        reporter, sender = self._new_reporter(request, batch_size=100, flush=.5)
        for i in range(50):
            reporter.report_span(self._new_span(str(i)))

        wait_for(lambda: len(sender.futures) > 0)
        assert 1 == len(sender.futures)
        sender.futures[0].set()

        for future in sender.futures[1:]:
            future.set()

        span_sent_key = 'jaeger:reporter_spans.result_ok'
        wait_for(lambda: span_sent_key in reporter.metrics_factory.counters)
        assert reporter.metrics_factory.counters[span_sent_key] == 50

    def test_sender_flush_errors_handled_with_partial_batch(self, request):
        reporter, sender = self._new_reporter(request, batch_size=100, flush=.5)
        reporter.error_reporter = ErrorReporter(
            metrics=Metrics(), logger=logging.getLogger())
        reporter._sender.send = mock.MagicMock(side_effect=ValueError())
        for i in range(50):
            reporter.report_span(self._new_span(str(i)))

        reporter_failure_key = 'jaeger:reporter_spans.result_err'
        wait_for(lambda: reporter_failure_key in reporter.metrics_factory.counters)
        assert reporter.metrics_factory.counters[reporter_failure_key] == 50

    def test_submit_failure(self, request):
        reporter, sender = self._new_reporter(request, batch_size=1)
        reporter.error_reporter = ErrorReporter(
            metrics=Metrics(), logger=logging.getLogger())

        reporter_failure_key = 'jaeger:reporter_spans.result_err'
        assert reporter_failure_key not in reporter.metrics_factory.counters

        reporter._sender.send = mock.MagicMock(side_effect=ValueError())
        reporter.report_span(self._new_span('1'))
        wait_for(lambda: reporter_failure_key in reporter.metrics_factory.counters)
        assert 1 == reporter.metrics_factory.counters.get(reporter_failure_key)

    def test_submit_queue_full_batch_size_1(self, request):
        reporter, sender = self._new_reporter(request, batch_size=1, queue_cap=1)
        reporter.report_span(self._new_span('1'))

        wait_for(lambda: len(sender.futures) > 0)
        assert 1 == len(sender.futures)
        # the consumer is blocked on a future, so won't drain the queue
        reporter.report_span(self._new_span('2'))
        span_dropped_key = 'jaeger:reporter_spans.result_dropped'
        assert span_dropped_key not in reporter.metrics_factory.counters
        reporter.report_span(self._new_span('3'))
        wait_for(lambda: span_dropped_key in reporter.metrics_factory.counters)
        assert 1 == reporter.metrics_factory.counters.get(span_dropped_key)
        # let it drain the queue
        sender.futures[0].set()
        wait_for(lambda: len(sender.futures) > 1)
        assert 2 == len(sender.futures)

        sender.futures[1].set()
        reporter.close()

    def test_submit_batch_size_2(self, request):
        reporter, sender = self._new_reporter(request, batch_size=2, flush=0.01)
        reporter.report_span(self._new_span('1'))
        time.sleep(0.001)
        assert 0 == len(sender.futures)

        reporter.report_span(self._new_span('2'))
        wait_for(lambda: len(sender.futures) > 0)
        assert 1 == len(sender.futures)
        assert 2 == len(sender.requests[0].spans)
        sender.futures[0].set()

        # 3rd span will not be submitted right away, but after `flush` interval
        reporter.report_span(self._new_span('3'))
        time.sleep(0.001)
        assert 1 == len(sender.futures)
        time.sleep(0.001)
        assert 1 == len(sender.futures)
        time.sleep(0.01)
        assert 2 == len(sender.futures)
        sender.futures[1].set()

        reporter.close()

    def test_close_drains_queue(self, request):
        reporter, sender = self._new_reporter(request, batch_size=1, flush=5)
        reporter.report_span(self._new_span('0'))

        wait_for(lambda: len(sender.futures) > 0)
        assert 1 == len(sender.futures)

        # now that the consumer is blocked on the first future.
        # let's reset Send to actually respond right away
        # and flood the queue with messages
        count = [0]

        def send(_):
            count[0] += 1
            return 1

        reporter._sender.send = send
        reporter._sender.batch_size = 3
        for i in range(10):
            reporter.report_span(self._new_span('%s' % i))
        assert reporter.queue.qsize() == 10, 'queued 10 spans'

        # now unblock consumer
        sender.futures[0].set()
        assert wait_for(lambda: count[0] > 2)

        assert count[0] == 3, '9 out of 10 spans submitted in 3 batches'
        assert wait_for(
            lambda: reporter.queue.unfinished_tasks == 1
        ), 'one span still pending'

        reporter.close()
        assert reporter.queue.qsize() == 0, 'all spans drained'
        assert count[0] == 4, 'last span submitted in one extrac batch'

    def test_flush_interval(self, request):
        reporter_store = [None]
        count = [0]

        def send(_):
            count[0] += 1
            if count[0] < 10:
                reporter_store[0].report_span(self._new_span('span'))
            return 1

        reporter, sender = self._new_reporter(request, batch_size=100, flush=.001, send_method=send)
        reporter_store[0] = reporter

        # initiate lazy thread creation
        reporter.report_span(self._new_span('span'))

        assert wait_for((lambda: count[0] == 10), ms=100)
        reporter.close()



# Copyright (c) 2016 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from jaeger_client import SpanContext


def test_parent_id_to_none():
    ctx1 = SpanContext(trace_id=1, span_id=2, parent_id=0, flags=1)
    assert ctx1.parent_id is None


def test_with_baggage_items():
    baggage1 = {'x': 'y'}
    ctx1 = SpanContext(trace_id=1, span_id=2, parent_id=3, flags=1,
                       baggage=baggage1)
    ctx2 = ctx1.with_baggage_item('a', 'b')
    assert ctx1.trace_id == ctx2.trace_id
    assert ctx1.span_id == ctx2.span_id
    assert ctx1.parent_id == ctx2.parent_id
    assert ctx1.flags == ctx2.flags
    assert ctx1.baggage != ctx2.baggage
    baggage1['a'] = 'b'
    assert ctx1.baggage == ctx2.baggage

    ctx3 = ctx2.with_baggage_item('a', None)
    assert ctx2.baggage != ctx3.baggage
    baggage1.pop('a')
    assert ctx3.baggage == baggage1


def test_is_debug_id_container_only():
    ctx = SpanContext.with_debug_id('value1')
    assert ctx.is_debug_id_container_only
    assert ctx.debug_id == 'value1'
    ctx = SpanContext(trace_id=1, span_id=2, parent_id=3, flags=1)
    assert not ctx.is_debug_id_container_only
